import functools
import logging
import typing as tp
from unittest.mock import AsyncMock, MagicMock

import pytest
import redis.asyncio as redis
from aio_pika import IncomingMessage

import protobunny as pb_base
from protobunny import RequeueMessage
from protobunny import asyncio as pb
from protobunny.asyncio.backends import mosquitto as mosquitto_backend_aio
from protobunny.asyncio.backends import python as python_backend_aio
from protobunny.asyncio.backends import rabbitmq as rabbitmq_backend_aio
from protobunny.asyncio.backends import redis as redis_backend_aio
from protobunny.backends import python as python_backend
from protobunny.config import backend_configs
from protobunny.helpers import (
    get_queue,
)
from protobunny.models import Envelope, IncomingMessageProtocol

from . import tests
from .utils import (
    assert_backend_connection,
    assert_backend_publish,
    assert_backend_setup_queue,
    async_wait,
    get_mocked_connection,
    incoming_message_factory,
)


@pytest.mark.parametrize(
    "backend", [rabbitmq_backend_aio, redis_backend_aio, python_backend_aio, mosquitto_backend_aio]
)
@pytest.mark.asyncio
class TestConnection:
    @pytest.fixture(autouse=True)
    async def mock_connections(
        self, backend, mocker, mock_redis_client, mock_aio_pika, mock_mosquitto, test_config
    ) -> tp.AsyncGenerator[dict[str, AsyncMock | None], None]:
        backend_name = backend.__name__.split(".")[-1]

        test_config.mode = "async"
        test_config.backend = backend_name
        test_config.log_task_in_redis = True
        test_config.backend_config = backend_configs[backend_name]
        mocker.patch.object(pb_base.config, "default_configuration", test_config)
        mocker.patch.object(pb_base.models, "default_configuration", test_config)
        mocker.patch.object(pb_base.backends, "default_configuration", test_config)
        mocker.patch.object(pb_base.helpers, "default_configuration", test_config)
        mocker.patch.object(pb.backends, "default_configuration", test_config)
        connection_module = getattr(pb.backends, backend_name).connection
        if hasattr(connection_module, "default_configuration"):
            mocker.patch.object(connection_module, "default_configuration", test_config)

        assert pb_base.helpers.get_backend() == backend
        assert pb.get_backend() == backend
        assert isinstance(get_queue(tests.tasks.TaskMessage), backend.queues.AsyncQueue)
        conn_with_fake_internal_conn = get_mocked_connection(
            backend, mock_redis_client, mock_aio_pika, mocker, mock_mosquitto
        )
        mocker.patch.object(pb, "connect", return_value=conn_with_fake_internal_conn)
        mocker.patch.object(pb, "disconnect", side_effect=backend.connection.disconnect)
        mocker.patch(
            "protobunny.asyncio.backends.BaseAsyncQueue.get_connection",
            return_value=conn_with_fake_internal_conn,
        )
        mocker.patch(
            f"protobunny.asyncio.backends.{backend_name}.connection.connect",
            return_value=conn_with_fake_internal_conn,
        )
        yield {
            "rabbitmq": mock_aio_pika,
            "redis": mock_redis_client,
            "connection": conn_with_fake_internal_conn,
            "python": conn_with_fake_internal_conn._connection,
            "mosquitto": mock_mosquitto,
        }
        connection_module.Connection.instance_by_vhost.clear()

    @pytest.fixture
    async def mock_connection(self, mock_connections, backend):
        yield mock_connections["connection"]

    @pytest.fixture
    async def mock_internal_connection(self, mock_connections, backend):
        backend_name = backend.__name__.split(".")[-1]
        yield mock_connections[backend_name]

    async def test_connection_success(
        self, mock_connection: MagicMock, mock_internal_connection, backend
    ) -> None:
        await mock_connection.connect()
        assert mock_connection.is_connected()
        await mock_connection.disconnect()
        assert not mock_connection.is_connected()
        await assert_backend_connection(backend, mock_internal_connection)

    async def test_setup_queue_shared(
        self, mock_connection: MagicMock, mock_internal_connection, backend
    ):
        await mock_connection.connect()
        await mock_connection.setup_queue("mylib.tasks.TaskMessage", shared=True)
        await assert_backend_setup_queue(
            backend,
            mock_internal_connection,
            "mylib.tasks.TaskMessage",
            shared=True,
            mock_connection=mock_connection,
        )

    async def test_publish_tasks(
        self, mock_connection: MagicMock, mock_internal_connection, backend
    ):
        conn = await mock_connection.connect()

        incoming = incoming_message_factory(backend)
        backend_name = backend.__name__.split(".")[-1]
        topic = "test.tasks.key"
        delimiter = backend_configs[backend_name].topic_delimiter
        topic = topic.replace(".", delimiter)

        await conn.publish(topic, incoming)
        # assert backend call
        await assert_backend_publish(
            backend,
            mock_internal_connection,
            mock_connection,
            incoming,
            topic=topic,
            count_in_queue=1,
            shared_queue=True,
        )

    async def test_publish(self, mock_connection: MagicMock, mock_internal_connection, backend):
        topic = "test.routing.key"
        conn = await mock_connection.connect()
        msg = None
        incoming = incoming_message_factory(backend)

        async def callback(envelope: IncomingMessageProtocol):
            nonlocal msg
            msg = envelope

        async def predicate():
            return msg == incoming

        await conn.subscribe(topic, callback=callback)
        await conn.publish(topic, incoming)
        if backend == python_backend:
            assert await async_wait(predicate)

        await assert_backend_publish(
            backend,
            mock_internal_connection,
            mock_connection,
            incoming,
            topic=topic,
            count_in_queue=0,
        )

    @pytest.mark.asyncio
    async def test_singleton_logic(self, backend):
        conn1 = await backend.connection.connect()
        conn2 = await backend.connection.connect()
        assert conn1 is conn2
        await conn1.disconnect()


# --- Specific Tests for rabbitmq ---


@pytest.mark.asyncio
async def test_on_message_requeue_rmq(mock_aio_pika):
    conn = rabbitmq_backend_aio.connection.Connection(requeue_delay=0)  # No delay for testing
    mock_msg = AsyncMock(spec=IncomingMessage)

    # Callback that triggers requeue
    def side_effect(*args):
        raise RequeueMessage()

    callback = MagicMock(side_effect=side_effect)

    await conn._on_message("test.topic", callback, mock_msg)

    mock_msg.reject.assert_awaited_once_with(requeue=True)


@pytest.mark.asyncio
async def test_on_message_poison_pill(mock_aio_pika):
    conn = rabbitmq_backend_aio.connection.Connection()
    mock_msg = AsyncMock(spec=IncomingMessage)

    # Random crash
    def crash(*args):
        raise RuntimeError("Boom")

    callback = MagicMock(side_effect=crash)

    await conn._on_message("test.topic", callback, mock_msg)

    # Should reject without requeue to avoid infinite loop
    mock_msg.reject.assert_awaited_once_with(requeue=False)


@pytest.mark.asyncio
async def test_on_message_success(mock_redis_client):
    conn = redis_backend_aio.connection.Connection()
    await conn.connect()
    # Mock an incoming message
    callback = AsyncMock()

    # We call the internal _on_message
    await conn._on_message_task(
        "test.tasks.topic",
        callback=callback,
        payload={"body": b"test"},
        group_name="test",
        msg_id="111",
    )
    assert callback.called


# --- Specific Tests for redis ---


@pytest.mark.asyncio
async def test_on_message_requeue_redis(mock_redis_client: redis.Redis):
    async with redis_backend_aio.connection.Connection() as conn:
        conn.requeue_delay = 0.1
        payload = Envelope(body=b"test")
        topic = "test.topic"

        # Callback that triggers requeue
        def side_effect(*args):
            raise RequeueMessage()

        msg = None
        callback = MagicMock(side_effect=side_effect)
        # compose the callback for the connection and call it
        real_cb = functools.partial(conn._on_message_pubsub, topic=topic, callback=callback)
        pubsub = mock_redis_client.pubsub()
        await pubsub.subscribe(f"protobunny:{topic}")
        await real_cb(envelope=payload)
        log = logging.getLogger(__name__)

        async def predicate() -> bool:
            nonlocal msg

            msg = await pubsub.get_message(ignore_subscribe_messages=True, timeout=1)
            return msg is not None

        # the message should be requeued and appeared in the queue
        assert await async_wait(predicate)
        assert msg["type"] in ["message", "pmessage"]
        assert msg["data"] == payload.body


@pytest.mark.asyncio
async def test_on_message_poison_pill(mock_redis_client, mocker):
    spy = mocker.spy(mock_redis_client, "publish")
    async with redis_backend_aio.connection.Connection() as conn:
        # Random crash
        def crash(*args):
            raise Exception("Boom")

        callback = MagicMock(side_effect=crash)

        with pytest.raises(Exception):
            await conn._on_message_pubsub("test.topic", callback=callback, body=b"test")

    # Should reject without requeue to avoid poisoning the queue
    spy.assert_not_called()


async def test_get_message_count(mock_redis_client):
    async with redis_backend_aio.connection.Connection() as conn:
        await conn.publish("test:tasks:topic", Envelope(body=b"test message 1"))
        await conn.publish("test:tasks:topic", Envelope(body=b"test message 2"))
        await conn.publish("test:tasks:topic", Envelope(body=b"test message 3"))
        count = await conn.get_message_count("test:tasks:topic")
        assert count == 3



import asyncio
import threading
import typing as tp
from collections import defaultdict

from .models import BaseQueue


class SubscriptionRegistry:
    def __init__(self):
        self.lock = asyncio.Lock()
        self.sync_lock = threading.Lock()
        self.subscriptions: dict[str, "BaseQueue"] = dict()
        self.results_subscriptions: dict[str, "BaseQueue"] = dict()
        self.tasks_subscriptions: dict[str, list["BaseQueue"]] = defaultdict(list)

    @staticmethod
    def get_key(message_type) -> str:
        return str(message_type)

    def register_subscription(self, message_type, queue):
        self.subscriptions[self.get_key(message_type)] = queue

    def register_results(self, message_type, queue):
        self.results_subscriptions[self.get_key(message_type)] = queue

    def register_task(self, message_type, queue):
        self.tasks_subscriptions[self.get_key(message_type)].append(queue)

    def unregister_subscription(self, message_type):
        self.subscriptions.pop(self.get_key(message_type), None)

    def unregister_results(self, message_type) -> "BaseQueue | None":
        return self.results_subscriptions.pop(self.get_key(message_type), None)

    def unregister_tasks(self, message_type):
        return self.tasks_subscriptions.pop(self.get_key(message_type), [])

    def get_subscription(self, message_type) -> "BaseQueue | None":
        return self.subscriptions.get(self.get_key(message_type), None)

    def get_tasks(self, message_type) -> list["BaseQueue"]:
        return self.tasks_subscriptions.get(self.get_key(message_type), [])

    def get_results(self, message_type) -> "BaseQueue | None":
        return self.results_subscriptions.get(self.get_key(message_type), None)

    def get_all_subscriptions(self) -> tp.ValuesView["BaseQueue"]:
        return self.subscriptions.values()

    def get_all_results(self) -> tp.ValuesView["BaseQueue"]:
        return self.results_subscriptions.values()

    def get_all_tasks(
        self, flat: bool = False
    ) -> tp.Generator["BaseQueue", None, None] | list["BaseQueue"]:
        if flat:
            for queues in self.tasks_subscriptions.values():
                for q in queues:
                    yield q
        else:
            yield from self.tasks_subscriptions.values()

    def unregister_all(self):
        self.subscriptions.clear()
        self.tasks_subscriptions.clear()
        self.results_subscriptions.clear()

    def unregister_all_tasks(self):
        self.tasks_subscriptions.clear()

    def unregister_all_subscriptions(self):
        self.subscriptions.clear()

    def unregister_all_results(self):
        self.results_subscriptions.clear()


# Then create a default instance
default_registry = SubscriptionRegistry()



import datetime
import json
import uuid
from decimal import Decimal
from typing import Any

numpy_installed = False
try:
    import numpy as np

    numpy_installed = True
except ImportError:
    pass


class ProtobunnyJsonEncoder(json.JSONEncoder):
    def default(self, obj: object) -> Any:
        if isinstance(obj, datetime.datetime):
            if obj.tzinfo is None:
                obj = obj.replace(tzinfo=datetime.timezone.utc)
            return obj.isoformat()
        elif isinstance(obj, datetime.date):
            return obj.isoformat()
        elif isinstance(obj, uuid.UUID):
            return str(obj)
        elif isinstance(obj, bytes):
            return str(obj)
        elif isinstance(obj, Decimal):
            return float(obj)

        if not numpy_installed:
            return super().default(obj)
        else:
            if isinstance(obj, np.ndarray):
                return obj.tolist()
            elif isinstance(obj, np.integer):
                return int(obj)
            elif isinstance(obj, np.floating):
                return float(obj)
        return super().default(obj)



import asyncio
import time
from types import ModuleType
from unittest.mock import ANY, AsyncMock

from aio_pika import Message
from redis import asyncio as redis
from waiting import TimeoutExpired, wait

from protobunny.backends import BaseConnection
from protobunny.models import Envelope


async def async_wait(condition_func, timeout=1.0, sleep=0.1) -> bool:
    start_time = time.time()
    while time.time() - start_time < timeout:
        res = await condition_func()
        if res:
            return True
        await asyncio.sleep(sleep)  # This yields control to the loop!
    return False


def sync_wait(condition_func, timeout=1.0, sleep=0.1) -> bool:
    try:
        wait(condition_func, timeout_seconds=timeout, sleep_seconds=sleep)
    except TimeoutExpired:
        return False
    return True


async def tear_down(event_loop):
    # Collect all tasks and cancel those that are not 'done'.
    tasks = asyncio.all_tasks(event_loop)
    tasks = [t for t in tasks if not t.done()]
    for task in tasks:
        task.cancel()
    # Wait for all tasks to complete, ignoring any CancelledErrors
    try:
        await asyncio.wait(tasks)
    except asyncio.exceptions.CancelledError:
        pass


def incoming_message_factory(backend, body: bytes = b"Hello"):
    backend_name = backend.__name__.split(".")[-1]
    if backend_name == "rabbitmq":
        return Message(body=body)
    elif backend_name == "redis":
        return Envelope(body=body, correlation_id="123")
    return Envelope(body=body)


async def assert_backend_publish(
    backend: ModuleType,
    internal_mock: dict | redis.Redis | AsyncMock,
    mock_connection: "BaseConnection",
    backend_msg,
    topic,
    count_in_queue: int = 1,
    shared_queue: bool = False,
):
    backend_name = backend.__name__.split(".")[-1]
    match backend_name:
        case "rabbitmq":
            internal_mock["exchange"].publish.assert_awaited_with(
                backend_msg, routing_key=topic, mandatory=True, immediate=False
            )

        case "redis":
            key = mock_connection.build_topic_key(topic)
            if shared_queue:
                # Read the last message from the stream
                incoming = await internal_mock.xread({key: 0})

                topic, messages = incoming[0]

                msg_id, msg = messages[0]
                assert (
                    len(messages) == count_in_queue
                ), f"Expected {count_in_queue} message in stream {key}, got {len(messages)}"
                assert (
                    msg[b"body"] == backend_msg.body
                ), f"Expected body {backend_msg.body}, got {msg[b'body']}"

            else:
                # retest the flow here
                pubsub = internal_mock.pubsub()
                await pubsub.subscribe(key)
                await mock_connection.publish(topic, backend_msg)
                msg: dict[str, bytes] | None = None

                async def predicate():
                    nonlocal msg
                    msg = await pubsub.get_message(timeout=1, ignore_subscribe_messages=True)

                    return msg is not None

                assert await async_wait(predicate)

                assert msg is not None
                assert (
                    msg["channel"].decode() == key
                ), f"Expected {topic}, got {msg['channel'].decode()}"
                assert (
                    msg["data"] == backend_msg.body
                ), f"Expected {backend_msg.body}, got {msg['data']}"

        case "python":
            assert internal_mock._exclusive_queues[topic] is not None
            assert (
                await internal_mock.get_message_count(topic) == count_in_queue
            ), f"count was {await internal_mock.get_message_count(topic)}"

        case "mosquitto":
            internal_mock.publish.assert_awaited_once()


async def assert_backend_setup_queue(
    backend, internal_mock, topic: str, shared: bool, mock_connection
) -> None:
    backend_name = backend.__name__.split(".")[-1]
    if backend_name == "rabbitmq":
        internal_mock["channel"].declare_queue.assert_called_with(
            topic, exclusive=not shared, durable=True, auto_delete=False, arguments=ANY
        )
    elif backend_name == "redis":
        if shared:
            streams = await internal_mock.xinfo_groups(f"protobunny:{topic}")
            assert len(streams) == 1
            assert (
                streams[0]["name"].decode() == "shared_group"
            ), f"Expected 'shared_group', got '{streams[0]['name']}'"
        else:
            assert mock_connection.queues[topic] == {"is_shared": False}

    elif backend_name == "python":
        if not shared:
            assert len(internal_mock._exclusive_queues.get(topic)) == 1
        else:
            assert internal_mock._shared_queues.get(topic)
    elif backend_name == "mosquitto":
        if not shared:
            assert mock_connection.queues[topic] == {
                "is_shared": False,
                "group_name": "",
                "sub_key": f"$share/shared_group/test/{topic}",
                "tag": ANY,
                "topic": topic,
                "topic_key": f"test/{topic}",
            }
        else:
            assert list(mock_connection.queues.values())[0] == {
                "is_shared": True,
                "group_name": "shared_group",
                "sub_key": f"$share/shared_group/protobunny/{topic}",
                "tag": ANY,
                "topic": topic,
                "topic_key": f"protobunny/{topic}",
            }, mock_connection.queues


async def assert_backend_connection(backend, internal_mock):
    backend_name = backend.__name__.split(".")[-1]
    if backend_name == "rabbitmq":
        # Verify aio_pika calls
        internal_mock["connect"].assert_awaited_once()
        assert internal_mock["channel"].set_qos.called
        # Check if main and DLX exchanges were declared
        assert internal_mock["channel"].declare_exchange.call_count == 2
    elif backend_name == "redis":
        assert await internal_mock.ping()
    elif backend_name == "mosquitto":
        internal_mock.__aenter__.assert_awaited_once()
    assert True


def get_mocked_connection(backend, redis_client, mock_aio_pika, mocker, mock_mosquitto):
    backend_name = backend.__name__.split(".")[-1]
    if backend_name == "redis":
        real_conn_with_fake_redis = backend.connection.Connection(url="redis://localhost:6379/0")
        assert (
            real_conn_with_fake_redis._exchange == "protobunny"
        ), real_conn_with_fake_redis._exchange
        real_conn_with_fake_redis._connection = redis_client

        def check_connected() -> bool:
            return real_conn_with_fake_redis._connection is not None

        # Patch is_connected logic
        mocker.patch.object(real_conn_with_fake_redis, "is_connected", side_effect=check_connected)

        return real_conn_with_fake_redis
    elif backend_name == "rabbitmq":
        real_conn_with_fake_aio_pika = backend.connection.Connection(
            url="amqp://guest:guest@localhost:5672/"
        )
        real_conn_with_fake_aio_pika._connection = mock_aio_pika["connection"]
        real_conn_with_fake_aio_pika.is_connected_event.set()
        real_conn_with_fake_aio_pika._channel = mock_aio_pika["channel"]
        real_conn_with_fake_aio_pika._exchange = mock_aio_pika["exchange"]
        real_conn_with_fake_aio_pika._queue = mock_aio_pika["queue"]

        return real_conn_with_fake_aio_pika
    elif backend_name == "python":
        python_conn = backend.connection.Connection()
        python_conn.is_connected_event.set()
        return python_conn
    elif backend_name == "mosquitto":
        real_conn_with_fake_aiomqtt = backend.connection.Connection()
        real_conn_with_fake_aiomqtt._connection = mock_mosquitto
        # real_conn_with_fake_aiomqtt.is_connected_event.set()
        return real_conn_with_fake_aiomqtt



from unittest.mock import MagicMock

import aio_pika
import pytest
from aio_pika import DeliveryMode
from pytest_mock import MockerFixture

import protobunny as pb
from protobunny.backends.rabbitmq.connection import Connection
from protobunny.config import backend_configs
from protobunny.models import (
    deserialize_result_message,
    get_message_class_from_topic,
    get_message_class_from_type_url,
)

from . import tests


@pytest.fixture(autouse=True)
def setup_connections(mocker: MockerFixture, mock_sync_rmq_connection, test_config) -> None:
    from protobunny.backends import rabbitmq as rabbitmq_backend

    test_config.mode = "sync"
    test_config.backend = "rabbitmq"
    test_config.backend_config = backend_configs["rabbitmq"]
    mocker.patch.object(pb.config, "default_configuration", test_config)
    mocker.patch.object(pb.models, "default_configuration", test_config)
    mocker.patch.object(pb.backends, "default_configuration", test_config)
    mocker.patch.object(pb.helpers, "default_configuration", test_config)
    pb.backend = rabbitmq_backend
    mocker.patch.object(pb, "connect", rabbitmq_backend.connection.connect)
    mocker.patch.object(pb.helpers.default_configuration, "backend", "rabbitmq")
    queue = pb.get_queue(tests.TestMessage)
    assert isinstance(queue, rabbitmq_backend.queues.SyncQueue)
    assert isinstance(queue.get_connection(), Connection)


def test_serdeser_result() -> None:
    msg = tests.TestMessage(content="test", number=123)
    result = msg.make_result(return_value={"test": "value"})
    assert result.source_message.value == bytes(msg)
    assert result.return_value == {"test": "value"}
    assert result.source == msg
    message_type = get_message_class_from_topic(msg.topic)
    assert isinstance(msg, message_type)

    # result.source_message is a protobuf.Any instance,
    # (and not an instance of the original Start message)
    # To get an instance of the source message,
    # use get_message_class_from_type_url and pass the Any.type_url as parameter
    message_type = get_message_class_from_type_url(result.source_message.type_url)
    assert isinstance(msg, message_type)
    assert msg == message_type().parse(bytes(msg))

    queue = pb.get_queue(msg)
    assert "acme.tests.TestMessage.result" == queue.result_topic

    # Serialize result
    # The return_value dictionary is first converted
    # to a commons.JsonContent before being serialized
    ser = bytes(result)
    deser = deserialize_result_message(ser)
    assert deser.source_message.type_url == "tests.tests.TestMessage"
    assert deser.source_message.type_url == msg.type_url
    # The return value is a commons.JsonContent object
    # and it transparently returns to a dict once deserialized
    assert deser.return_value == {"test": "value"}
    assert deser.error == "" and deser.error == result.error
    assert deser == result


def test_topics(mock_sync_rmq_connection: MagicMock) -> None:
    msg = tests.TestMessage(content="test", number=123)
    result = msg.make_result(return_value={"test": "value"})
    q = pb.get_queue(result.source)
    assert isinstance(q.get_connection(), Connection)
    assert q.get_connection() == mock_sync_rmq_connection
    assert q.result_topic == "acme.tests.TestMessage.result"
    pb.publish_result(result)
    expected_payload = aio_pika.Message(
        bytes(result),
        correlation_id=None,
        delivery_mode=DeliveryMode.NOT_PERSISTENT,
    )
    mock_sync_rmq_connection.publish.assert_called_once_with(
        "acme.tests.TestMessage.result", expected_payload
    )



import logging

from protobunny.asyncio.backends import (
    BaseAsyncQueue,
)

log = logging.getLogger(__name__)


class AsyncQueue(BaseAsyncQueue):
    def get_tag(self) -> str:
        return self.subscription



import logging
import os
import threading
import typing as tp
import uuid
from abc import ABC
from collections import defaultdict
from concurrent.futures import ThreadPoolExecutor
from queue import Empty, Queue

from ... import RequeueMessage
from ...config import default_configuration
from ...models import Envelope, SyncCallback, is_task
from .. import BaseConnection

log = logging.getLogger(__name__)

VHOST = os.environ.get("PYTHON_VHOST", "/")


class MessageBroker:
    def __init__(self):
        self._shared_queues: dict[str, list[Queue]] = defaultdict(list)
        self._exclusive_queues: dict[str, list[Queue]] = defaultdict(list)
        self.lock = threading.RLock()
        self.logger_queue: Queue | None = None

    def publish(self, topic: str, message: Envelope) -> bool:
        """Publish a message to all relevant queues."""
        published = False
        shared = is_task(topic)
        with self.lock:
            if self.logger_queue:
                self.logger_queue.put(message)
            if shared:
                queues = self._shared_queues.get(topic)
                if not queues:
                    log.warning("No subscribers for tasks %s", topic)
                    return False
                # Select the queue with the least messages (Load Balancing)
                best_queue = min(queues, key=lambda q: q.qsize())
                best_queue.put(message)
                return True
            # For exact matches:
            if topic in self._exclusive_queues:
                for queue in self._exclusive_queues[topic]:
                    queue.put(message)
                    published = True
            # Fan out
            for sub_topic, queues in self._exclusive_queues.items():
                if sub_topic.endswith("#") and topic.startswith(sub_topic[:-1]):
                    for queue in queues:
                        queue.put(message)
                        published = True

        return published

    def create_shared_queue(self, topic: str) -> Queue:
        """Get or create a shared queue."""
        with self.lock:
            queue = Queue()
            self._shared_queues[topic].append(queue)
            return queue

    def create_exclusive_queue(self, topic: str) -> Queue:
        """Create an exclusive queue for a topic."""
        with self.lock:
            queue = Queue()
            self._exclusive_queues[topic].append(queue)
            return queue

    def remove_exclusive_queue(self, topic: str, queue: Queue) -> None:
        """Remove an exclusive queue."""
        with self.lock:
            if topic in self._exclusive_queues:
                try:
                    self._exclusive_queues[topic].remove(queue)
                except ValueError:
                    pass

    def remove_shared_queues(self, topic: str) -> None:
        """Remove all in process subscriptions for a shared queue."""
        with self.lock:
            self._shared_queues.pop(topic, None)

    def purge_queue(self, topic: str) -> None:
        """Empty a shared queue."""
        with self.lock:
            queues = self._shared_queues.get(topic, [])
            for queue in queues:
                while not queue.empty():
                    try:
                        queue.get_nowait()
                    except Empty:
                        break

    def get_message_count(self, topic: str) -> int:
        """Get queue size of the shared queues."""
        with self.lock:
            return (
                sum(queue.qsize() for queue in self._shared_queues.get(topic, []))
                if topic in self._shared_queues
                else 0
            )


class BaseLocalConnection(BaseConnection, ABC):
    """Base class with shared logic for python "connections"."""

    instance_by_vhost: dict[str, tp.Any]
    lock: tp.Any
    _connection: MessageBroker

    def __init__(self, vhost: str = "/", requeue_delay: int = 3):
        self.vhost = vhost
        self.requeue_delay = requeue_delay
        self._is_connected = False
        self._subscriptions: dict[str, dict] = {}
        self.logger_prefix = default_configuration.logger_prefix


class Connection(BaseLocalConnection):
    """Synchronous local connection using threads."""

    instance_by_vhost: dict[str, "Connection"] = {}
    _is_connected_event: threading.Event | None
    lock = threading.RLock()
    _connection_cls = MessageBroker

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._connection = self._connection_cls()
        self._is_connected_event = None
        self.executor = ThreadPoolExecutor(max_workers=10, thread_name_prefix="PB Py-Worker")

    @staticmethod
    def create_tag(topic: str, shared: bool) -> str:
        """Generate subscription tag."""
        context_id = threading.get_ident() if not shared else uuid.uuid4()
        suffix = f"shared-{context_id}" if shared else context_id
        return f"local-sub-{topic}-{suffix}"

    def purge(self, topic: str, **kwargs) -> None:
        self._connection.purge_queue(topic)

    def get_message_count(self, topic: str) -> int:
        return self._connection.get_message_count(topic)

    def setup_queue(self, topic: str, shared: bool) -> Queue:
        """Create appropriate queue type."""
        if topic == self.logger_prefix:
            self._connection.logger_queue = Queue()
            return self._connection.logger_queue
        elif shared:
            return self._connection.create_shared_queue(topic)
        else:
            return self._connection.create_exclusive_queue(topic)

    def is_connected(self) -> bool:
        return self.is_connected_event.is_set()

    @property
    def is_connected_event(self) -> threading.Event:
        """Lazily create the event in the current running loop."""
        if self._is_connected_event is None:
            self._is_connected_event = threading.Event()
        return self._is_connected_event

    @classmethod
    def get_connection(cls, vhost: str = "/") -> "Connection":
        if vhost not in cls.instance_by_vhost:
            with cls.lock:
                if vhost not in cls.instance_by_vhost:
                    instance = cls(vhost=vhost)
                    instance.connect()
                    cls.instance_by_vhost[vhost] = instance
        return cls.instance_by_vhost[vhost]

    def connect(self, timeout: float = 10.0) -> "Connection":
        with self.lock:
            self.is_connected_event.set()
            return self

    def disconnect(self, **kwargs) -> None:
        with self.lock:
            log.info("Disconnecting SyncLocalConnection for vhost: %s", self.vhost)
            for tag in list(self._subscriptions.keys()):
                self._unsubscribe_by_tag(tag, if_unused=False)
            # Shutdown the pool
            self.executor.shutdown(wait=True)
            self.is_connected_event.clear()
            self.instance_by_vhost.pop(self.vhost, None)

    def publish(self, topic: str, message: Envelope, **kwargs) -> None:
        if not self._connection.publish(topic, message):
            log.warning("No subscribers for topic '%s'", topic)

    def _message_worker(
        self, topic: str, callback: SyncCallback, stop_event: threading.Event, queue: Queue
    ) -> None:
        """Worker thread for processing messages."""
        while not stop_event.is_set():
            message = None
            try:
                message = queue.get(timeout=0.5)
                if stop_event.is_set():
                    break
                callback(message)
                queue.task_done()
                log.debug("Message processed on topic '%s'", topic)
            except RequeueMessage:
                log.warning("Requeuing message on topic '%s'", topic)
                # Start a timer to requeue so this thread isn't blocked
                threading.Timer(self.requeue_delay, lambda m=message: queue.put(m)).start()
                queue.task_done()
            except Empty:
                continue
            except Exception:
                log.exception("Error processing message on topic '%s'", topic)

    def subscribe(self, topic: str, callback: SyncCallback, shared: bool = False) -> str:
        with self.lock:
            tag = self.create_tag(topic, shared)
            if tag in self._subscriptions:
                log.warning("Already subscribed with tag '%s'", tag)
                return tag

            queue = self.setup_queue(topic, shared)
            stop_event = threading.Event()
            future = self.executor.submit(self._message_worker, topic, callback, stop_event, queue)

            self._subscriptions[tag] = {
                "topic": topic,
                "shared": shared,
                "queue": queue,
                "stop_event": stop_event,
                "future": future,
            }

            log.info("Subscribed to topic '%s' (tag: %s)", topic, tag)
            return tag

    def unsubscribe(self, topic: str, if_unused: bool = True, **kwargs) -> None:
        with self.lock:
            tags = [tag for tag, info in self._subscriptions.items() if info["topic"] == topic]
            for tag in tags:
                self._unsubscribe_by_tag(tag, if_unused)

    def _cleanup_queue(self, topic: str, queue: Queue, shared: bool, if_unused: bool) -> None:
        """Clean up queue resources."""
        if topic == self.logger_prefix:
            self._connection.logger_queue = None
        elif not shared:
            self._connection.remove_exclusive_queue(topic, queue)
        elif not if_unused:
            self._connection.remove_shared_queues(topic)

    def _unsubscribe_by_tag(self, tag: str, if_unused: bool = True) -> None:
        """Internal unsubscribe by tag."""
        log.debug("Unsubscribing by tag: %s", tag)
        info = self._subscriptions.pop(tag, None)
        if not info:
            return

        log.info("Unsubscribing from topic '%s' (tag: %s)", info["topic"], tag)

        # Stop worker thread
        info["stop_event"].set()
        # Cleanup queue
        self._cleanup_queue(info["topic"], info["queue"], info["shared"], if_unused)

    def get_consumer_count(self, topic: str) -> int:
        with self.lock:
            return sum(1 for info in self._subscriptions.values() if info["topic"] == topic)

    def __enter__(self):
        self.connect()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.disconnect()
        return False


# Convenience functions
def connect() -> Connection:
    return Connection.get_connection(vhost=VHOST)


def reset_connection() -> Connection:
    connection = connect()
    connection.disconnect()
    return connect()


def disconnect() -> None:
    connection = connect()
    connection.disconnect()



"""A script that add imports for submodules in betterproto messages' packages.

It allows code assistance in IDEs.
>>> import mylib as ml
>>> msg = ml.machine.control.MoveUp()
Note: It must be executed after code generation/post install script.
- It's included in make command `compile` and in setup.py for lib installations.
- It's part of the protobunny compile command
"""
import argparse
import black
import logging
import os
import re
from argparse import Namespace
from collections import defaultdict
from pathlib import Path
from typing import Any, TextIO

from jinja2 import Environment, FileSystemLoader

logging.basicConfig(level=logging.INFO)
log = logging.getLogger("protobunny")


def get_package_path(package_name: str, base_dir: str) -> str:
    """Give the package path given a package name and base directory, return the package path.
    Raise ImportError if package is not found.

    Args:
        package_name: The package name.
        base_dir: The base directory.
    Returns:
        The package path.
    Raises:
        ImportError: If package is not found.
    """
    package_path = os.path.join(base_dir, package_name.replace(".", os.sep))
    if os.path.isdir(package_path) and os.path.exists(os.path.join(package_path, "__init__.py")):
        log.info("Found package %s", package_path)
        return package_path
    raise ImportError(f"Package {package_name} not found in {base_dir}")


def get_modules_with_subpackages(package_name: str, base_dir: str = "./") -> dict[str, list[str]]:
    """Walks the directory corresponding to a package (located in base_dir) for subpackages,
       and returns a dictionary mapping package names to a list of their subpackage names,
       useful to determine the imports to add later.

    Args:
        package_name: The root package name for the generated python code e.g. 'mylib.codegen'
        base_dir: The base directory e.g. '/home/john/projects/myproject'

    Returns:
        A dictionary mapping package names to a list of their subpackage names.
    """
    res = defaultdict(list)
    try:
        package_path = get_package_path(package_name, base_dir)
    except ImportError:
        # Exit from recursion
        return res

    # Iterate through the items in the package directory.
    for item in os.listdir(package_path):
        item_path = os.path.join(package_path, item)
        # Check if the item is a directory and has an __init__.py file.
        if os.path.isdir(item_path) and os.path.exists(os.path.join(item_path, "__init__.py")):
            res[package_name].append(item)
            full_name = f"{package_name}.{item}"
            # Recursively update the dictionary with subpackages.
            res.update(get_modules_with_subpackages(full_name, base_dir))
    return res


def post_compile_step_1(pkg: str, subpackages: list[str], source_dir: str = "./") -> None:
    init_path = Path(source_dir) / pkg.replace(".", os.sep) / "__init__.py"
    with init_path.open("a") as init_file:
        # Append subpackages imports to the __init__.py file
        write_header(init_file, breaklines=2)
        for subpackage in subpackages:
            init_file.write(f"from . import {subpackage.split('.')[-1]}  # noqa\n")


def post_compile_step_2(pkg: str, source_dir: str = "./") -> None:
    init_path = Path(source_dir) / pkg.replace(".", os.sep) / "__init__.py"
    source = init_path.read_text()
    source = ensure_message_mixin(source)
    source = ensure_dict_type(source)
    source = black.format_str(
        src_contents=source,
        mode=black.Mode(),
    )
    init_path.write_text(source)


def write_main_init(main_imports: list[str], main_package: str, source_dir: str = "./") -> None:
    """
    Imports main packages into the `__init__.py` file at the specified source directory. This function
    handles rendering the `__init__.py` file using a Jinja2 template if available, formats the content
    using the Black code formatter, and writes the final output to the file.

    Args:
        main_imports (list[str]): A list of main imports to be included in the `__init__.py` file.
        main_package (str): The full dotted path of the main package associated with the imports.
        source_dir (str): The directory where the top-level package is located. Defaults to `"./"`.
    """
    # Import main packages in __init__.py
    # find the root package for the messaging library
    packages = main_package.split(".")
    top_level_package, generated_package_name = (packages[-1], "") if len(packages) == 1 else (packages[0], packages[1])
    path = Path(source_dir) / top_level_package
    if not path.exists():
        return
    # Write the main __init__.py files for the protobunny (both sync and async)
    init_paths = [path / "asyncio" / "__init__.py", path / "__init__.py"]
    # look for a jinja template in the main directory
    for init_path in init_paths:
        if (init_path.parent / "__init__.py.j2").exists():
            write_from_template(generated_package_name, init_path, main_imports, init_path.parent)
        elif init_path.parent.exists():
            # Write the init of the generated library with just the import
            with open(init_path, mode="w+") as fh:
                write_header(fh)
                for subpackage in main_imports:
                    fh.write(f"from .{generated_package_name} import {subpackage.split('.')[-1]}  # noqa\n")


def write_from_template(generated_package_name: str, init_path: Path, main_imports: list[str], path: Path):
    environment = Environment(loader=FileSystemLoader(path))
    template = environment.get_template("__init__.py.j2")
    # Render the templates with the main imports
    main_init_source = template.render(
        generated_package_name=generated_package_name, main_imports=main_imports
    )
    main_init_source = black.format_str(
        src_contents=main_init_source,
        mode=black.Mode(),
    )
    with open(init_path, mode="w+") as fh:
        fh.write(main_init_source)


def write_header(fh: TextIO, breaklines: int = 0):
    fh.write("\n" * breaklines)
    fh.write("#" * 55)
    fh.write("\n# Dynamically added by protobunny post_compile.py\n\n")


def replace_typing_import(source: str, new_import: str) -> str:
    pattern = re.compile(
        r"^(\s*)from\s+typing\s+import\s*"
        r"(?:"
        r"([^\(].*?)$"  # single-line: from typing import X, Y
        r"|"
        r"\(\s*\n(.+?)\n\s*\)"  # multiline: from typing import (\n    X,\n    Y\n)
        r")",
        re.MULTILINE | re.DOTALL,
    )

    def repl(match):
        indent = match.group(1)  # preserve original indentation
        return indent + new_import.replace("\n", "\n" + indent)

    return pattern.sub(repl, source, count=1)


def ensure_dict_type(source: str) -> str:
    """Add Dict type annotation to JsonContent fields.

    Args:
        source: the python source file intended for modification.
    """
    # Pattern 1: Optional["_commons__.JsonContent"] or
    # Matches: options: Optional["_commons__.JsonContent"] = betterproto.message_field(
    pattern1 = re.compile(
        r'(\w+):\s*Optional\["(_*commons__\.JsonContent)"\]\s*=\s*betterproto\.message_field\(',
        re.MULTILINE,
    )
    # Pattern 2: _commons__.JsonContent (without Optional)
    # Matches: options: _commons__.JsonContent = betterproto.message_field(
    pattern2 = re.compile(
        r"(\w+):\s*\"(_*commons__\.JsonContent)\"\s*=\s*betterproto\.message_field\(",
        re.MULTILINE,
    )

    def replace1(match):
        field_name = match.group(1)
        type_name = match.group(2)
        return f'{field_name}: Optional["{type_name} | dict"] = betterproto.message_field('

    def replace2(match):
        field_name = match.group(1)
        type_name = match.group(2)
        return f"{field_name}: {type_name} | dict = betterproto.message_field("
    new_source = pattern1.sub(replace1, source)
    new_source = pattern2.sub(replace2, new_source)
    return new_source


def ensure_message_mixin(source: str) -> str:
    new_source = source.replace("(betterproto.Message):", "(models.ProtoBunnyMessage):")
    lines = new_source.split('\n')

    for i, line in enumerate(lines):
        if 'import betterproto' in line:
            # Insert after this line
            lines.insert(i + 1, "from protobunny import models as models")
            break

    return '\n'.join(lines)


def main() -> None:
    args = cli_args()
    main_package = args.proto_pkg
    source_dir = args.source_dir
    packages = get_modules_with_subpackages(main_package, source_dir)
    all_packages = get_all_submodules(packages)
    log.debug("Packages found to have submodules: %s", packages)
    log.info("All packages: %s", all_packages)
    main_imports = packages.pop(main_package)
    # Write main protobunny init file using jinja template,
    # to use the full namespace syntax
    # In [1]: import mylib as ml
    # In [2]: ml.main.MyMessage
    # Out[3]: mylib.codegen.main.MyMessage
    write_main_init(main_imports, main_package, source_dir)
    # For each of the main subpackages, check if it has subpackages.
    # If so, we append imports to the __init__.py file
    for pkg in packages:
        post_compile_step_1(pkg, packages[pkg], source_dir)
    # For all subpackages
    # 1 - add MixIn to all Message classes
    # 2 - ensure that JsonContent fields are typed as Union[JsonContent, Dict]
    for pkg in all_packages:
        log.info("Type hinting package %s", pkg)
        post_compile_step_2(pkg, source_dir)


def get_all_submodules(packages: dict[str, list[str]]) -> set[Any]:
    all_packages = set()
    for pkg_name, subpackages in packages.items():
        for subpackage in subpackages:
            full_name = f"{pkg_name}.{subpackage}"
            all_packages.add(full_name)
    return all_packages


def cli_args() -> Namespace:
    parser = argparse.ArgumentParser(
        description="""
         protobunny post compile:
           add imports for subpackages
           add dict type annotation to JsonContent
        """
    )
    parser.add_argument(
        "-s",
        "--source-dir",
        type=str,
        help="Base source directory",
        required=False,
        default="./",
    )
    parser.add_argument(
        "-p",
        "--proto-pkg",
        type=str,
        required=False,
        default="codegen",
        help="betterproto code generated root package (i.e. codegen)",
    )
    args = parser.parse_args()
    return args


if __name__ == "__main__":
    main()



import typing as tp
from unittest.mock import AsyncMock, MagicMock, patch

import aio_pika
import aiormq
import fakeredis
import pamqp
import pytest
import pytest_asyncio
from pytest_mock import MockerFixture

import protobunny
from protobunny.asyncio.backends.mosquitto import connection as mosquitto_connection_aio
from protobunny.asyncio.backends.python import connection as python_connection_aio
from protobunny.asyncio.backends.rabbitmq import connection as rabbitmq_connection_aio
from protobunny.asyncio.backends.redis import connection as redis_connection_aio
from protobunny.backends.mosquitto import connection as mosquitto_connection
from protobunny.backends.python import connection as python_connection
from protobunny.backends.rabbitmq import connection as rabbitmq_connection
from protobunny.backends.redis import connection as redis_connection


@pytest_asyncio.fixture
def test_config() -> protobunny.config.Config:
    conf = protobunny.config.Config(
        messages_directory="tests/proto",
        messages_prefix="acme",
        generated_package_name="tests",
        project_name="test",
        project_root="./",
        force_required_fields=True,
        mode="async",
        backend="rabbitmq",
        log_task_in_redis=False,
    )
    return conf


class MockMQTTConnection:
    mock_message = MagicMock()
    mock_message.payload = b"hello world"
    mock_message.topic.value = "test/topic"

    def __init__(self):
        self.subscribe = AsyncMock()
        self.unsubscribe = AsyncMock()
        self.publish = AsyncMock()

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc, tb):
        pass

    async def messages(self):
        for item in [self.mock_message]:
            yield item


@pytest.fixture
async def mock_mosquitto(mocker) -> tp.AsyncGenerator[MockMQTTConnection, None]:
    mock_client = MockMQTTConnection()
    mocker.spy(mock_client, "publish")
    mocker.spy(mock_client, "subscribe")
    mocker.spy(mock_client, "unsubscribe")
    mocker.spy(mock_client, "__aenter__")
    mocker.spy(mock_client, "__aexit__")

    with patch("aiomqtt.Client") as mock_client_class:
        mock_client_class.return_value = mock_client
        yield mock_client


@pytest.fixture
async def mock_redis_client(mocker) -> tp.AsyncGenerator[fakeredis.FakeAsyncRedis, None]:
    server = fakeredis.FakeServer()
    client = fakeredis.FakeAsyncRedis(decode_responses=False)

    mocker.patch("protobunny.asyncio.backends.redis.connection.redis.from_url", return_value=client)
    # Clear database before each test
    await client.flushdb()
    yield client
    # Explicitly clear internal fakeredis blocking listeners
    server.connected = False
    await client.aclose()


@pytest.fixture
async def mock_aio_pika():
    """Mocks the entire aio_pika connection chain."""
    with patch("aio_pika.connect_robust", new_callable=AsyncMock) as mock_connect:
        # Mock Connection
        mock_conn = AsyncMock()
        mock_connect.return_value = mock_conn

        # Mock Channel
        mock_channel = AsyncMock()
        mock_conn.channel.return_value = mock_channel

        # Mock Exchange
        mock_exchange = AsyncMock()
        mock_channel.declare_exchange.return_value = mock_exchange

        # Mock Queue
        mock_queue = AsyncMock()
        mock_queue.name = "test-queue"
        mock_queue.exclusive = False
        mock_channel.declare_queue.return_value = mock_queue

        yield {
            "connect": mock_connect,
            "connection": mock_conn,
            "channel": mock_channel,
            "exchange": mock_exchange,
            "queue": mock_queue,
        }


@pytest.fixture
def mock_sync_rmq_connection(mocker: MockerFixture) -> tp.Generator[MagicMock, None, None]:
    mock = mocker.MagicMock(spec=rabbitmq_connection.Connection)
    mocker.patch("protobunny.backends.BaseSyncQueue.get_connection", return_value=mock)
    mocker.patch("protobunny.backends.rabbitmq.connection.connect", return_value=mock)
    yield mock


@pytest.fixture
async def mock_rmq_connection(mocker: MockerFixture) -> tp.AsyncGenerator[AsyncMock, None]:
    mock = mocker.AsyncMock(spec=rabbitmq_connection_aio.Connection)
    mocker.patch("protobunny.backends.BaseAsyncQueue.get_connection", return_value=mock)
    yield mock


@pytest.fixture
def mock_sync_redis_connection(mocker: MockerFixture) -> tp.Generator[MagicMock, None, None]:
    mock = mocker.MagicMock(spec=redis_connection.Connection)
    mocker.patch("protobunny.backends.BaseSyncQueue.get_connection", return_value=mock)
    mocker.patch("protobunny.backends.rabbitmq.connection.connect", return_value=mock)
    yield mock


@pytest.fixture
async def mock_redis_connection(mocker: MockerFixture) -> tp.AsyncGenerator[AsyncMock, None]:
    mock = mocker.AsyncMock(spec=redis_connection_aio.Connection)
    mocker.patch("protobunny.backends.BaseAsyncQueue.get_connection", return_value=mock)
    yield mock


@pytest.fixture
def mock_sync_mqtt_connection(mocker: MockerFixture) -> tp.Generator[MagicMock, None, None]:
    mock = mocker.MagicMock(spec=mosquitto_connection.Connection)
    mocker.patch("protobunny.backends.BaseSyncQueue.get_connection", return_value=mock)
    mocker.patch("protobunny.backends.mosquitto.connection.connect", return_value=mock)
    yield mock


@pytest.fixture
async def mock_mqtt_connection(mocker: MockerFixture) -> tp.AsyncGenerator[AsyncMock, None]:
    mock = mocker.AsyncMock(spec=mosquitto_connection_aio.Connection)
    mocker.patch("protobunny.backends.BaseAsyncQueue.get_connection", return_value=mock)
    yield mock


@pytest.fixture
def mock_sync_python_connection(mocker: MockerFixture) -> tp.Generator[MagicMock, None, None]:
    mock = mocker.MagicMock(spec=python_connection.Connection)
    mocker.patch("protobunny.helpers.get_backend", return_value=protobunny.backends.python)
    mocker.patch("protobunny.backends.BaseSyncQueue.get_connection", return_value=mock)
    yield mock


@pytest.fixture
async def mock_python_connection(mocker: MockerFixture) -> tp.AsyncGenerator[AsyncMock, None]:
    mock = mocker.AsyncMock(spec=python_connection_aio.Connection)
    mocker.patch("protobunny.asyncio.backends.BaseAsyncQueue.get_connection", return_value=mock)
    mocker.patch("protobunny.helpers.get_backend", return_value=protobunny.backends.python)
    yield mock


@pytest.fixture
def pika_incoming_message() -> tp.Callable[[bytes, str], aio_pika.IncomingMessage]:
    def _incoming_message_factory(body: bytes, routing_key: str) -> aio_pika.IncomingMessage:
        return aio_pika.IncomingMessage(
            aiormq.abc.DeliveredMessage(
                header=pamqp.header.ContentHeader(),
                body=body,
                delivery=pamqp.commands.Basic.Deliver(routing_key=routing_key),
                channel=None,
            )
        )

    return _incoming_message_factory


@pytest.fixture
def pika_message() -> (
    tp.Callable[
        [
            bytes,
        ],
        aio_pika.Message,
    ]
):
    def _message_factory(body: bytes) -> aio_pika.Message:
        return aio_pika.Message(body=body)

    return _message_factory


@pytest.fixture(scope="session", autouse=True)
def pika_messages_eq() -> tp.Generator[None, None, None]:
    # Add support for equality in pika Messages
    # as the mock library uses args comparison for expected calls
    # and aio_pika.Message doesn't have __eq__ defined
    def compare_aio_pika_messages(a, b) -> bool:
        if not (isinstance(a, aio_pika.Message) and isinstance(b, aio_pika.Message)):
            return False
        return str(a) == str(b) and a.body == b.body

    aio_pika.Message.__eq__ = compare_aio_pika_messages  # type: ignore
    yield
    aio_pika.Message.__eq__ = object.__eq__  # type: ignore


@pytest.fixture(autouse=True)
def clear_cached_func() -> tp.Generator[None, None, None]:
    from protobunny.helpers import _build_routing_key

    _build_routing_key.cache_clear()
    yield
    _build_routing_key.cache_clear()



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: protobunny/results.proto
# plugin: python-betterproto
# This file has been @generated

from dataclasses import dataclass
from typing import Optional

import betterproto
import betterproto.lib.google.protobuf as betterproto_lib_google_protobuf

from protobunny import models as models

from .. import commons as _commons__


class ReturnCode(betterproto.Enum):
    MSG_FAILURE = 0
    FAILURE = 1
    IN_PROGRESS = 2
    SUCCESS = 3
    PENDING = 4


@dataclass(eq=False, repr=False)
class Result(models.ProtoBunnyMessage):
    source_message: "betterproto_lib_google_protobuf.Any" = betterproto.message_field(1)
    """
    * The source message. It is transparently deserialized by the `source` property
    ex.
    >>> msg = ml.mypackage.MyMessage(content="test")
    >>> result = msg.make_result()
    >>> assert msg == result.source
    """

    return_code: "ReturnCode" = betterproto.enum_field(2)
    error: Optional[str] = betterproto.string_field(3, optional=True)
    return_value: Optional["_commons__.JsonContent | dict"] = betterproto.message_field(
        4, optional=True
    )



"""Implements a RabbitMQ Connection with both sync and async support using aio_pika."""
import asyncio
import logging
import os
import threading

from ...asyncio.backends.rabbitmq.connection import Connection as RabbitMQConnection
from .. import BaseSyncConnection

log = logging.getLogger(__name__)

VHOST = os.environ.get("RABBITMQ_VHOST", "/")


def connect() -> "Connection":
    """Get the singleton async connection."""
    conn = Connection.get_connection(vhost=VHOST)
    return conn


def reset_connection() -> "Connection":
    """Reset the singleton connection."""
    connection = connect()
    connection.disconnect()
    return connect()


def disconnect() -> None:
    connection = connect()
    connection.disconnect()


class Connection(BaseSyncConnection):
    """Synchronous wrapper around Async Rmq Connection.

    Manages a dedicated event loop in a background thread to run async operations.

    Example:
        .. code-block:: python

            with Connection() as conn:
                conn.publish("my.topic", message)
                tag = conn.subscribe("my.topic", callback)

    """

    _lock = threading.RLock()
    _stopped: asyncio.Event | None = None
    instance_by_vhost: dict[str, "Connection"] = {}
    async_class = RabbitMQConnection



import asyncio
import logging
import time
import typing as tp

import pytest

import protobunny as pb_sync
from protobunny import asyncio as pb
from protobunny.asyncio.backends import mosquitto as mosquitto_backend_aio
from protobunny.asyncio.backends import python as python_backend_aio
from protobunny.asyncio.backends import rabbitmq as rabbitmq_backend_aio
from protobunny.asyncio.backends import redis as redis_backend_aio
from protobunny.backends import mosquitto as mosquitto_backend
from protobunny.backends import python as python_backend
from protobunny.backends import rabbitmq as rabbitmq_backend
from protobunny.backends import redis as redis_backend
from protobunny.config import backend_configs
from protobunny.models import ProtoBunnyMessage

from . import tests
from .utils import async_wait, sync_wait

log = logging.getLogger(__name__)


@pytest.mark.integration
@pytest.mark.asyncio
@pytest.mark.parametrize(
    "backend", [rabbitmq_backend_aio, redis_backend_aio, python_backend_aio, mosquitto_backend_aio]
)
class TestTasks:
    msg = tests.tasks.TaskMessage(content="test", bbox=[1, 2, 3, 4])
    received = {
        "task_1": None,
        "task_2": None,
    }

    @pytest.fixture(autouse=True)
    async def setup_test_env(self, mocker, test_config, backend) -> tp.AsyncGenerator[None, None]:
        backend_name = backend.__name__.split(".")[-1]
        test_config.mode = "async"
        test_config.backend = backend_name
        test_config.backend_config = backend_configs[backend_name]
        self.topic_delimiter = test_config.backend_config.topic_delimiter
        # Patch global configuration for all modules that use it
        mocker.patch.object(pb_sync.config, "default_configuration", test_config)
        mocker.patch.object(pb_sync.models, "default_configuration", test_config)
        mocker.patch.object(pb_sync.helpers, "default_configuration", test_config)
        mocker.patch.object(pb_sync.backends, "default_configuration", test_config)
        mocker.patch.object(pb.backends, "default_configuration", test_config)
        if hasattr(backend.connection, "default_configuration"):
            mocker.patch.object(backend.connection, "default_configuration", test_config)
        if hasattr(backend.queues, "default_configuration"):
            mocker.patch.object(backend.queues, "default_configuration", test_config)

        pb.backend = backend
        mocker.patch("protobunny.helpers.get_backend", return_value=backend)
        mocker.patch.object(pb, "connect", backend.connection.connect)
        mocker.patch.object(pb, "disconnect", backend.connection.disconnect)
        mocker.patch.object(pb, "get_backend", return_value=backend)

        # Assert the patching is working for setting the backend
        connection = await pb.connect()
        assert isinstance(connection, backend.connection.Connection)
        queue = pb.get_queue(self.msg)
        assert queue.topic == "acme.tests.tasks.TaskMessage".replace(
            ".", test_config.backend_config.topic_delimiter
        )
        assert isinstance(queue, backend.queues.AsyncQueue)
        assert isinstance(await queue.get_connection(), backend.connection.Connection)
        await queue.purge(reset_groups=True)
        # start without pending subscriptions
        await pb.unsubscribe_all(if_unused=False, if_empty=False)
        # reset the variables holding the messages received
        self.received = {}
        yield

        await connection.disconnect()
        backend.connection.Connection.instance_by_vhost.clear()

    async def test_tasks(self, backend) -> None:
        async def predicate_1() -> bool:
            return self.received.get("task_1") is not None

        async def predicate_2() -> bool:
            return self.received.get("task_2") is not None

        async def callback_task_1(msg: "ProtoBunnyMessage") -> None:
            log.debug("CALLBACK TASK 1 %s", msg)
            await asyncio.sleep(0.1)  # simulate some work
            self.received["task_1"] = msg

        async def callback_task_2(msg: "ProtoBunnyMessage") -> None:
            log.debug("CALLBACK TASK 2 %s", msg)
            await asyncio.sleep(0.1)  # simulate some work
            self.received["task_2"] = msg

        await pb.subscribe(tests.tasks.TaskMessage, callback_task_1)
        await pb.subscribe(tests.tasks.TaskMessage, callback_task_2)
        await pb.publish(self.msg)
        assert await async_wait(predicate_1)

        assert self.received.get("task_2") is None
        self.received["task_1"] = None
        await pb.publish(self.msg)
        await pb.publish(self.msg)
        await pb.publish(self.msg)
        assert await async_wait(predicate_1, timeout=2, sleep=0.1)
        assert await async_wait(predicate_2)
        assert self.received["task_1"] == self.msg
        assert self.received["task_2"] == self.msg
        self.received["task_1"] = None
        self.received["task_2"] = None
        await pb.publish(self.msg)
        await pb.publish(self.msg)
        assert await async_wait(predicate_1)
        assert await async_wait(predicate_2)
        await pb.unsubscribe(tests.tasks.TaskMessage, if_unused=False, if_empty=False)


@pytest.mark.integration
@pytest.mark.parametrize(
    "backend", [rabbitmq_backend, redis_backend, python_backend, mosquitto_backend]
)
class TestTasksSync:
    msg = tests.tasks.TaskMessage(content="test", bbox=[1, 2, 3, 4])
    received = {
        "task_1": None,
        "task_2": None,
    }

    @pytest.fixture(autouse=True)
    def setup_test_env(self, mocker, test_config, backend) -> tp.Generator[None, None, None]:
        backend_name = backend.__name__.split(".")[-1]
        test_config.mode = "sync"
        test_config.backend = backend_name
        test_config.backend_config = backend_configs[backend_name]
        self.topic_delimiter = test_config.backend_config.topic_delimiter
        # Patch global configuration for all modules that use it
        mocker.patch.object(pb_sync.config, "default_configuration", test_config)
        mocker.patch.object(pb_sync.models, "default_configuration", test_config)
        mocker.patch.object(pb_sync.backends, "default_configuration", test_config)
        mocker.patch.object(pb_sync.helpers, "default_configuration", test_config)
        mocker.patch.object(pb.backends.redis.connection, "default_configuration", test_config)
        if hasattr(backend.connection, "default_configuration"):
            mocker.patch.object(backend.connection, "default_configuration", test_config)
        if hasattr(backend.queues, "default_configuration"):
            mocker.patch.object(backend.queues, "default_configuration", test_config)

        pb_sync.backend = backend
        mocker.patch("protobunny.backends.get_backend", return_value=backend)
        mocker.patch("protobunny.helpers.get_backend", return_value=backend)
        mocker.patch.object(pb_sync, "connect", backend.connection.connect)
        mocker.patch.object(pb_sync, "disconnect", backend.connection.disconnect)

        # Assert the patching is working for setting the backend
        connection = pb_sync.connect()
        assert isinstance(connection, backend.connection.Connection)
        assert sync_wait(connection.is_connected)
        task_queue = pb_sync.get_queue(self.msg)
        assert task_queue.topic == "acme.tests.tasks.TaskMessage".replace(
            ".", test_config.backend_config.topic_delimiter
        )
        assert isinstance(task_queue, backend.queues.SyncQueue)
        assert isinstance(task_queue.get_connection(), backend.connection.Connection)
        task_queue.purge(reset_groups=True)
        # start without pending subscriptions
        pb_sync.unsubscribe_all(if_unused=False, if_empty=False)
        pb_sync.disconnect()
        # reset the variables holding the messages received
        self.received = {}

        yield

        pb_sync.disconnect()
        backend.connection.Connection.instance_by_vhost.clear()

    def test_tasks(self, backend) -> None:
        def predicate_1() -> bool:
            return self.received.get("task_1") is not None

        def predicate_2() -> bool:
            return self.received.get("task_2") is not None

        def callback_task_1(msg: "ProtoBunnyMessage") -> None:
            log.debug("SYNC CALLBACK TASK 1 %s", msg)
            time.sleep(0.1)
            self.received["task_1"] = msg

        def callback_task_2(msg: "ProtoBunnyMessage") -> None:
            log.debug("SYNC CALLBACK TASK 2 %s", msg)
            time.sleep(0.1)
            self.received["task_2"] = msg

        pb_sync.subscribe(tests.tasks.TaskMessage, callback_task_1)
        pb_sync.subscribe(tests.tasks.TaskMessage, callback_task_2)

        pb_sync.publish(self.msg)
        assert sync_wait(predicate_1)

        assert self.received.get("task_2") is None
        self.received["task_1"] = None
        pb_sync.publish(self.msg)
        pb_sync.publish(self.msg)
        pb_sync.publish(self.msg)
        assert sync_wait(predicate_1)
        assert sync_wait(predicate_2)
        assert self.received["task_1"] == self.msg
        assert self.received["task_2"] == self.msg
        self.received["task_1"] = None
        self.received["task_2"] = None
        pb_sync.publish(self.msg)
        pb_sync.publish(self.msg)
        assert sync_wait(predicate_1)
        assert sync_wait(predicate_2)
        self.received["task_2"] = None
        self.received["task_1"] = None



from unittest import mock

import pytest

from protobunny.config import load_config


@pytest.fixture(autouse=True)
def clear_config_cache():
    """Clear the cache before each test since load_config is cached."""
    load_config.cache_clear()
    yield
    load_config.cache_clear()


def test_config_default_values():
    """Test that default values are applied when no config files exist."""
    with mock.patch("pathlib.Path.exists", return_value=False):
        config = load_config()
        assert config.messages_prefix == "pb"
        assert config.backend == "rabbitmq"
        assert config.mode == "async"


def test_config_from_env(monkeypatch):
    """Test that environment variables override everything else."""
    monkeypatch.setenv("PROTOBUNNY_BACKEND", "redis")
    monkeypatch.setenv("PROTOBUNNY_MODE", "async")
    monkeypatch.setenv("PROTOBUNNY_FORCE_REQUIRED_FIELDS", "true")

    # Mocking pyproject and ini to be missing
    with mock.patch("pathlib.Path.exists", return_value=False):
        config = load_config()
        assert config.backend == "redis"
        assert config.mode == "async"
        assert config.force_required_fields is True


def test_config_from_ini(tmp_path, monkeypatch):
    """Test reading from protobunny.ini."""
    monkeypatch.chdir(tmp_path)
    ini_content = """
[protobunny]
backend = python
mode = async
force-required-fields = yes
"""
    (tmp_path / "protobunny.ini").write_text(ini_content)

    # Ensure pyproject doesn't interfere
    with mock.patch("protobunny.config.get_config_from_pyproject", return_value={}):
        config = load_config()
        assert config.backend == "python"
        assert config.mode == "async"
        assert config.force_required_fields is True


def test_config_precedence(tmp_path, monkeypatch):
    """Test that ENV > INI > pyproject."""
    monkeypatch.chdir(tmp_path)

    # 1. Setup INI
    ini_file = tmp_path / "protobunny.ini"
    ini_file.write_text("[protobunny]\nbackend = python\nmessages_prefix = ini_pref")

    # 2. Setup ENV (should override INI)
    monkeypatch.setenv("PROTOBUNNY_BACKEND", "redis")

    # Mock pyproject to provide a base
    mock_pyproject = {"backend": "rabbitmq", "messages_prefix": "py_pref", "project-name": "test"}

    with mock.patch("protobunny.config.get_config_from_pyproject", return_value=mock_pyproject):
        config = load_config()

        # Overridden by ENV
        assert config.backend == "redis"
        # Overridden by INI
        assert config.messages_prefix == "ini_pref"
        # Taken from pyproject
        assert config.project_name == "test"



import asyncio
import logging
import os
import typing as tp
import uuid
from abc import ABC
from collections import defaultdict
from queue import Empty, Queue

from ....config import default_configuration
from ....models import AsyncCallback, Envelope
from ... import RequeueMessage
from .. import BaseConnection, is_task

log = logging.getLogger(__name__)

VHOST = os.environ.get("PYTHON_VHOST", "/")


class MessageBroker:
    """Centralized message broker"""

    def __init__(self):
        self._shared_queues: dict[str, list[asyncio.Queue]] = defaultdict(list)
        self._exclusive_queues: dict[str, list[asyncio.Queue]] = defaultdict(list)
        self.lock = asyncio.Lock()
        self.logger_queue: asyncio.Queue | None = None

    async def publish(self, topic: str, message: Envelope) -> bool:
        """Publish a message to all relevant queues."""
        published = False
        shared = is_task(topic)
        async with self.lock:
            # Logger queue
            if self.logger_queue:
                await self.logger_queue.put(message)
            # Shared queue
            if shared:
                queues = self._shared_queues.get(topic)
                if not queues:
                    # A task message will create a shared queue if not existing yet
                    queue = asyncio.Queue()
                    self._shared_queues[topic].append(queue)
                    log.warning("No subscribers for tasks %s. Creating new queue", topic)
                    await queue.put(message)
                    return True
                # Select the queue with the least messages (Load Balancing)
                best_queue = min(queues, key=lambda q: q.qsize())
                await best_queue.put(message)
                return True

            # For exact matches:
            if topic in self._exclusive_queues:
                for queue in self._exclusive_queues[topic]:
                    await queue.put(message)
                    published = True

            # Fan out
            for sub_topic, queues in self._exclusive_queues.items():
                if sub_topic.endswith("#") and topic.startswith(sub_topic[:-1]):
                    for queue in queues:
                        await queue.put(message)
                        published = True

        return published

    async def create_shared_queue(self, topic: str) -> asyncio.Queue:
        """Get or create a shared queue."""
        async with self.lock:
            queue = asyncio.Queue()
            self._shared_queues[topic].append(queue)
            return queue

    async def create_exclusive_queue(self, topic: str) -> asyncio.Queue:
        """Create an exclusive queue for a topic."""
        async with self.lock:
            queue = asyncio.Queue()
            self._exclusive_queues[topic].append(queue)
            return queue

    async def remove_exclusive_queue(self, topic: str, queue: Queue) -> None:
        """Remove an exclusive queue."""
        async with self.lock:
            if topic in self._exclusive_queues:
                try:
                    self._exclusive_queues[topic].remove(queue)
                except ValueError:
                    pass

    async def remove_shared_queues(self, topic: str) -> None:
        """Remove all in process subscriptions for a shared queue."""
        async with self.lock:
            self._shared_queues.pop(topic, None)

    async def purge_queue(self, topic: str) -> None:
        """Empty a shared queue."""
        async with self.lock:
            queues = self._shared_queues.get(topic, [])
            for queue in queues:
                while not queue.empty():
                    try:
                        queue.get_nowait()
                    except Empty:
                        break

    async def get_message_count(self, topic: str) -> int:
        """Get queue size of the shared queues."""
        async with self.lock:
            return (
                sum(queue.qsize() for queue in self._shared_queues.get(topic, []))
                if topic in self._shared_queues
                else 0
            )


class BaseLocalConnection(BaseConnection, ABC):
    """Base class with shared logic for python "connections"."""

    instance_by_vhost: dict[str, tp.Any]
    lock: tp.Any
    _connection: MessageBroker

    def __init__(self, vhost: str = "/", requeue_delay: int = 3):
        self.vhost = vhost
        self.requeue_delay = requeue_delay
        self._is_connected = False
        self._subscriptions: dict[str, dict] = {}
        self.logger_prefix = default_configuration.logger_prefix


class Connection(BaseLocalConnection):
    """Asynchronous local connection using asyncio."""

    _connection_cls = MessageBroker
    instance_by_vhost: dict[str, "Connection"] = {}
    lock = asyncio.Lock()
    _is_connected_event: asyncio.Event | None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._connection = self._connection_cls()
        self._is_connected_event = None

    @staticmethod
    def create_tag(topic: str, shared: bool) -> str:
        """Generate subscription tag."""
        context_id = id(asyncio.current_task()) if not shared else uuid.uuid4()
        suffix = f"shared-{context_id}" if shared else context_id
        return f"local-sub-{topic}-{suffix}"

    async def setup_queue(self, topic: str, shared: bool = False) -> asyncio.Queue:
        """Create appropriate queue type."""
        if topic == self.logger_prefix:
            self._connection.logger_queue = asyncio.Queue()
            return self._connection.logger_queue
        elif shared:
            return await self._connection.create_shared_queue(topic)
        else:
            return await self._connection.create_exclusive_queue(topic)

    @property
    def is_connected_event(self) -> asyncio.Event:
        """Lazily create the event in the current running loop."""
        if self._is_connected_event is None:
            self._is_connected_event = asyncio.Event()
        return self._is_connected_event

    @classmethod
    async def get_connection(cls, vhost: str = "/") -> "Connection":
        if vhost not in cls.instance_by_vhost:
            async with cls.lock:
                if vhost not in cls.instance_by_vhost:
                    instance = cls(vhost=vhost)
                    await instance.connect()
                    cls.instance_by_vhost[vhost] = instance
        return cls.instance_by_vhost[vhost]

    def is_connected(self) -> bool:
        """Check if connection is established"""
        return self.is_connected_event.is_set()

    async def connect(self, timeout: float = 10.0) -> "Connection":
        log.info("Connecting async connection for vhost: %s", self.vhost)
        self.is_connected_event.set()
        return self

    async def disconnect(self, **kwargs) -> None:
        log.info("Disconnecting AsyncLocalConnection for vhost: %s", self.vhost)
        tags = list(self._subscriptions.keys())
        # Cancel all tasks first
        for tag in tags:
            info = self._subscriptions.get(tag)
            if info and "task" in info:
                info["stop_event"].set()
                info["task"].cancel()

        # Wait for cancellation to complete
        tasks = [info["task"] for info in self._subscriptions.values() if "task" in info]
        if tasks:
            await asyncio.gather(*tasks, return_exceptions=True)

        for tag in tags:
            await self._unsubscribe_by_tag(tag, if_unused=False)

        async with self.lock:
            self.is_connected_event.clear()
            self.instance_by_vhost.pop(self.vhost, None)

    async def publish(self, topic: str, message: Envelope, **kwargs) -> None:
        # Broker's publish is thread-safe
        published = await self._connection.publish(topic, message)
        if not published:
            log.warning("No subscribers for topic '%s'", topic)

    async def _message_worker(
        self, topic: str, callback: AsyncCallback, stop_event: asyncio.Event, queue: asyncio.Queue
    ) -> None:
        """Async worker for processing messages."""
        message = None
        while not stop_event.is_set():
            try:
                # Use thread pool for blocking queue.get
                message = await queue.get()
                if stop_event.is_set():
                    break
                await callback(message)
                queue.task_done()
                log.debug("Message processed on topic '%s'", topic)
            except RequeueMessage:
                log.warning("Requeuing message on topic '%s'", topic)
                await asyncio.sleep(self.requeue_delay)
                await queue.put(message)
                queue.task_done()
            except Empty:
                continue
            except asyncio.CancelledError:
                log.debug("Worker cancelled for topic '%s'", topic)
                raise
            except RuntimeError as e:
                if "Event loop is closed" in str(e):
                    log.debug("Worker stopped for topic '%s'", topic)
                    break
            except Exception:
                log.exception("Error processing message on topic '%s'", topic)
                continue

    async def subscribe(self, topic: str, callback: AsyncCallback, shared: bool = False) -> str:
        async with self.lock:
            tag = self.create_tag(topic, shared)

            if tag in self._subscriptions:
                log.warning("Already subscribed with tag '%s'", tag)
                return tag

            queue = await self.setup_queue(topic, shared)
            stop_event = asyncio.Event()

            task = asyncio.create_task(self._message_worker(topic, callback, stop_event, queue))

            self._subscriptions[tag] = {
                "topic": topic,
                "shared": shared,
                "queue": queue,
                "stop_event": stop_event,
                "task": task,
            }

            log.info("Subscribed to topic '%s' (tag: %s)", topic, tag)
            return tag

    async def unsubscribe(self, topic: str, if_unused: bool = True, if_empty: bool = True) -> None:
        async with self.lock:
            tags = [tag for tag, info in self._subscriptions.items() if info["topic"] == topic]
            for tag in tags:
                await self._unsubscribe_by_tag(tag, if_unused)

    async def _unsubscribe_by_tag(self, tag: str, if_unused: bool = True) -> None:
        """Internal unsubscribe by tag."""
        info = self._subscriptions.pop(tag, None)
        if not info:
            return

        log.info("Unsubscribing from topic '%s' (tag: %s)", info["topic"], tag)

        # Stop worker task
        info["stop_event"].set()
        task = info.get("task")
        if task and not task.done():
            task.cancel()
            try:
                # Await the task to ensure it's cleaned up from the event loop
                await asyncio.gather(task, return_exceptions=True)
            except asyncio.CancelledError:
                pass
        try:
            await asyncio.wait_for(task, timeout=1.0)
        except asyncio.TimeoutError:
            pass
        except asyncio.CancelledError:
            pass
        await self._cleanup_queue(info["topic"], info["queue"], info["shared"], if_unused)

    async def _cleanup_queue(
        self, topic: str, queue: asyncio.Queue, shared: bool, if_unused: bool
    ) -> None:
        """Clean up queue resources."""
        if topic == self.logger_prefix:
            self._connection.logger_queue = None
        elif not shared:
            await self._connection.remove_exclusive_queue(topic, queue)
        elif not if_unused:
            await self._connection.remove_shared_queues(topic)

    async def get_consumer_count(self, topic: str) -> int:
        async with self.lock:
            return sum(1 for info in self._subscriptions.values() if info["topic"] == topic)

    async def purge(self, topic: str, **kwargs) -> None:
        await self._connection.purge_queue(topic)

    async def get_message_count(self, topic: str) -> int:
        return await self._connection.get_message_count(topic)

    async def __aenter__(self):
        await self.connect()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.disconnect()


# Convenience functions
async def connect() -> Connection:
    return await Connection.get_connection(vhost=VHOST)


async def reset_connection() -> Connection:
    connection = await connect()
    await connection.disconnect()
    return await connect()


async def disconnect() -> None:
    connection = await connect()
    await connection.disconnect()



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: tasks.proto
# plugin: python-betterproto
# This file has been @generated

from dataclasses import dataclass
from typing import (
    List,
    Optional,
)

import betterproto

from protobunny import models as models

from ... import commons as __commons__


@dataclass(eq=False, repr=False)
class TaskMessage(models.ProtoBunnyMessage):
    content: str = betterproto.string_field(10)
    weights: List[float] = betterproto.float_field(30)
    bbox: List[int] = betterproto.int64_field(40)
    options: Optional["__commons__.JsonContent | dict"] = betterproto.message_field(
        50, optional=True
    )



"""Implements a Redis Connection with both sync and async support"""
import asyncio
import logging
import os
import threading

from ...asyncio.backends.redis.connection import Connection as RedisConnection
from .. import BaseSyncConnection

log = logging.getLogger(__name__)


VHOST = os.environ.get("REDIS_VHOST") or os.environ.get("REDIS_DB", "0")


def connect() -> "Connection":
    """Get the singleton async connection."""
    conn = Connection.get_connection(vhost=VHOST)
    return conn


def reset_connection() -> "Connection":
    """Reset the singleton connection."""
    connection = connect()
    connection.disconnect()
    return connect()


def disconnect() -> None:
    connection = connect()
    connection.disconnect()


class Connection(BaseSyncConnection):
    """Synchronous wrapper around the async connection

    Example:
        .. code-block:: python

            with SyncRedisConnection() as conn:
                conn.publish("my.topic", message)
                tag = conn.subscribe("my.topic", callback)

    """

    _lock = threading.RLock()
    _stopped: asyncio.Event | None = None
    instance_by_vhost: dict[str, "Connection"] = {}
    async_class = RedisConnection

    def reset_stream_groups(self, stream_key: str) -> None:
        self._run_coro(self._async_conn.reset_stream_groups(stream_key))



import configparser
import dataclasses
import functools
import inspect
import logging
import os
import typing as tp
from collections import namedtuple

try:
    import tomllib
except ImportError:
    import tomli as tomllib  # type: ignore

from pathlib import Path

PROJECT_NAME = "protobunny"
PACKAGE_NAME = "protobunny"
GENERATED_PACKAGE_NAME = "core"
ROOT_GENERATED_PACKAGE_NAME = f"{PACKAGE_NAME}.{GENERATED_PACKAGE_NAME}"
PREFIX_MESSAGES = "pb"
MESSAGES_DIRECTORY = "protobuf/protobunny"

ENV_PREFIX = "PROTOBUNNY_"
INI_FILE = "protobunny.ini"

AvailableBackends = tp.Literal["rabbitmq", "python", "redis", "mosquitto"]

log = logging.getLogger(__name__)


BackEndConfig = namedtuple(
    "BackedConfig", "topic_delimiter multi_wildcard_delimiter namespace", defaults=(None,) * 3
)

rabbitmq_backend_config = BackEndConfig(
    topic_delimiter=".", multi_wildcard_delimiter="#", namespace=""
)
python_backend_config = BackEndConfig(
    topic_delimiter=".", multi_wildcard_delimiter="#", namespace=""
)
redis_backend_config = BackEndConfig(
    topic_delimiter=":", multi_wildcard_delimiter="*", namespace="protobunny"
)
mosquitto_backend_config = BackEndConfig(
    topic_delimiter="/", multi_wildcard_delimiter="#", namespace="protobunny"
)

backend_configs = {
    "rabbitmq": rabbitmq_backend_config,
    "python": python_backend_config,
    "redis": redis_backend_config,
    "mosquitto": mosquitto_backend_config,
}


@dataclasses.dataclass
class Config:
    messages_directory: str = "messages"
    messages_prefix: str = PREFIX_MESSAGES
    project_name: str = PROJECT_NAME
    project_root: str = "./"
    force_required_fields: bool = False
    generated_package_name: str = "codegen"
    generated_package_root: str = "./"
    mode: tp.Literal["sync", "async"] = "async"
    backend: "AvailableBackends" = "rabbitmq"
    backend_config: BackEndConfig = rabbitmq_backend_config
    log_task_in_redis: bool = False
    available_backends = ("rabbitmq", "python", "redis", "mosquitto")

    def __post_init__(self) -> None:
        if self.mode not in ("sync", "async"):
            raise ValueError(f"Invalid mode: {self.mode}. Must be one of: sync, async")
        if not self.backend:
            raise ValueError("Backend not configured.")
        self.backend_config = backend_configs.get(self.backend)

    @property
    def use_async(self) -> bool:
        return self.mode == "async"

    @functools.cached_property
    def logger_prefix(self) -> str:
        return f"{self.messages_prefix}{self.backend_config.topic_delimiter}{self.backend_config.multi_wildcard_delimiter}"


@functools.cache
def load_config() -> Config:
    """Load user config
    1. `[tool.protobunny]` from the nearest pyproject.toml.
    2. `protobunny.ini` in the current directory.
    3. Environment variables prefixed with PROTOBUNNY_.
    """
    config_dict = get_config_from_pyproject()

    ini_config = get_config_from_ini()
    config_dict.update(ini_config)

    env_config = get_config_from_env()
    config_dict.update(env_config)

    if "generated-package-name" not in config_dict:
        config_dict["generated-package-name"] = (
            f"{config_dict['project-name']}.codegen" if "project-name" in config_dict else "codegen"
        )
    available_params = inspect.signature(Config).parameters.keys()
    passed_params = {k.replace("-", "_") for k in config_dict.keys()}
    unknown_params = passed_params - set(available_params)
    if unknown_params:
        log.warning(
            f"Config params: [{', '.join(unknown_params)}] are not handled in protobunny version {VERSION}. "
            f"Ignoring. Available params: {', '.join(available_params)}"
        )
    return Config(
        **{
            k.replace("-", "_"): v
            for k, v in config_dict.items()
            if k.replace("-", "_") in available_params
        }
    )


def get_config_from_env() -> dict[str, tp.Any]:
    """Read config from environment variables prefixed with PROTOBUNNY_."""
    config = {}
    # Map of config keys to their expected types for casting
    type_hints = {
        "force_required_fields": bool,
    }
    for key, value in os.environ.items():
        if not key.startswith(ENV_PREFIX):
            continue
        config_key = key[len(ENV_PREFIX) :].lower().replace("_", "-")
        # Basic type casting for booleans
        if type_hints.get(config_key.replace("-", "_")) is bool:
            config[config_key] = value.lower() in ("true", "1", "yes", "on")
        else:
            config[config_key] = value
    return config


def _get_pyproject_path() -> Path | None:
    start_path = Path.cwd()
    for folder in [start_path, *start_path.parents]:
        pyproject = folder / "pyproject.toml"
        if pyproject.exists():
            return pyproject


def get_config_from_pyproject() -> dict[str, tp.Any]:
    config = dict()
    pyproject = _get_pyproject_path()
    if not pyproject:
        return config
    data = tomllib.loads(pyproject.read_text())
    config = data.get("tool", {}).get("protobunny", {})
    config["project-root"] = pyproject.parent.name
    config["project-name"] = data["project"].get("name", PROJECT_NAME).replace("-", "_")
    return config


def get_config_from_ini() -> dict[str, tp.Any]:
    """Read config from protobunny.ini in the current directory."""
    path = Path.cwd() / INI_FILE
    if not path.exists():
        return {}

    parser = configparser.ConfigParser()
    parser.read(path)
    if "protobunny" not in parser:
        return {}

    config = {}
    for key, value in parser["protobunny"].items():
        # Handle boolean casting for specific keys
        if key == "force-required-fields":
            config[key] = parser["protobunny"].getboolean(key)
        else:
            config[key] = value
    return config


# Get project version from pyproject.toml for releasing
def get_project_version() -> str:
    """Read version from pyproject.toml in the project root."""
    pyproject = _get_pyproject_path()
    if not pyproject:
        return "0.0.0-dev"
    data = tomllib.loads(pyproject.read_text())
    return data.get("project", {}).get("version", "0.0.0-dev")


VERSION = get_project_version()
default_configuration = load_config()



from unittest.mock import ANY, MagicMock

import aio_pika
import pytest
from aio_pika import DeliveryMode

import protobunny as pb

from . import tests


@pytest.fixture(autouse=True)
def setup_config(mocker, test_config) -> None:
    test_config.mode = "sync"
    mocker.patch.object(pb.config, "default_configuration", test_config)
    mocker.patch.object(pb.models, "default_configuration", test_config)
    mocker.patch.object(pb.backends, "default_configuration", test_config)
    mocker.patch.object(pb.helpers, "default_configuration", test_config)


def test_sync_send_message(mock_sync_rmq_connection: MagicMock) -> None:
    msg = tests.TestMessage(content="test", number=123, color=tests.Color.GREEN)
    queue = pb.get_queue(msg, backend_name="rabbitmq")
    queue.publish(msg)
    expected_payload = aio_pika.Message(
        bytes(msg),
        correlation_id=None,
        delivery_mode=DeliveryMode.PERSISTENT,
    )
    mock_sync_rmq_connection.publish.assert_called_once_with(
        "acme.tests.TestMessage", expected_payload
    )


def test_sync_subscribe(mock_sync_rmq_connection: MagicMock) -> None:
    msg = tests.TestMessage()
    func = lambda x: print(x)  # noqa: E731
    pb.subscribe(msg, func)
    mock_sync_rmq_connection.subscribe.assert_called_with(
        "acme.tests.TestMessage", ANY, shared=False
    )
    pb.subscribe(tests.tasks.TaskMessage, func)
    mock_sync_rmq_connection.subscribe.assert_called_with(
        "acme.tests.tasks.TaskMessage", ANY, shared=True
    )
    pb.subscribe_results(tests.tasks.TaskMessage, func)
    mock_sync_rmq_connection.subscribe.assert_called_with(
        "acme.tests.tasks.TaskMessage.result", ANY, shared=False
    )



import functools
import importlib
import sys
import typing
from types import ModuleType

import betterproto

from .config import default_configuration

if typing.TYPE_CHECKING:
    from .asyncio.backends import BaseAsyncQueue
    from .backends import BaseSyncQueue
    from .models import ProtoBunnyMessage


def get_topic(pkg_or_msg: "ProtoBunnyMessage | type[ProtoBunnyMessage] | ModuleType") -> str:
    """Return a Topic dataclass object based on a Message (instance or class) or a ModuleType.

    It uses build_routing_key to determine the topic name.
    Note: The topic name can be a routing key with a binding key

    Args:
        pkg_or_msg: a Message instance, a Message class or a module

    Returns: topic string
    """
    delimiter = default_configuration.backend_config.topic_delimiter
    return f"{default_configuration.messages_prefix}{delimiter}{build_routing_key(pkg_or_msg)}"


def get_backend(backend: str | None = None) -> ModuleType:
    """
    Retrieve and import the specified backend module.

    Load the backend module based on the provided name or falls back
    to the default backend specified in the configuration. If the backend is unavailable
    or cannot be imported, it exits the program.

    Args:
        backend (str | None): The name of the backend to import. If None, the backend from
            the configuration is used.

    Returns:
        The imported backend module.
    """
    backend = backend or default_configuration.backend
    module = ".asyncio" if default_configuration.use_async else ""
    module_name = f"protobunny{module}.backends.{backend}"
    if module_name in sys.modules:
        return sys.modules[module_name]
    try:
        module = importlib.import_module(module_name)
    except ModuleNotFoundError as exc:
        suggestion = ""
        if backend not in default_configuration.available_backends:
            suggestion = f" Invalid backend or backend not supported.\nAvailable backends: {default_configuration.available_backends}"
        else:
            suggestion = (
                f" Install the backend with pip install protobunny[{backend}]."
                if backend != "python"
                else suggestion
            )
        sys.exit(f"Could not import backend: {exc}.{suggestion}")
    return module


def get_queue(
    pkg_or_msg: "ProtoBunnyMessage | type['ProtoBunnyMessage'] | ModuleType",
    backend_name: str | None = None,
) -> "BaseSyncQueue|BaseAsyncQueue":
    """Factory method to get an AsyncQueue/SyncQueue instance based on
      - the message type (e.g. mylib.subpackage.subsubpackage.MyMessage)
      - the mode (e.g. async)
      - the configured backend or the parameter passed (e.g. "rabbitmq")

    Args:
        pkg_or_msg: A message instance, a message class, or a module
            containing message definitions.
        backend_name: backend name to use

    Returns:
        Async/SyncQueue: A queue instance configured for the relevant topic.
    """
    backend_name = backend_name or default_configuration.backend
    queue_type = "AsyncQueue" if default_configuration.use_async else "SyncQueue"
    return getattr(get_backend(backend=backend_name).queues, queue_type)(get_topic(pkg_or_msg))


@functools.lru_cache(maxsize=100)
def _build_routing_key(module: str, cls_name: str) -> str:
    # Build the routing key from the module and class name
    backend = default_configuration.backend_config
    delimiter = backend.topic_delimiter
    routing_key = f"{module}.{cls_name}"
    config = default_configuration
    if not routing_key.startswith(config.generated_package_name):
        raise ValueError(
            f"Invalid topic {routing_key}, must start with {config.generated_package_name}."
        )
    # As convention, we set the topic name to the message class name,
    # left-stripped of the root generated package name
    # (e.g. my_messaging_lib.codegen.vision.control.Start => vision.control.Start)
    routing_key = routing_key.split(f"{config.generated_package_name}.", maxsplit=1)[1].replace(
        ".", delimiter
    )
    return routing_key


def build_routing_key(
    pkg_or_msg: "ProtoBunnyMessage | type[ProtoBunnyMessage] | ModuleType",
) -> str:
    """Returns a routing key based on a message instance, a message class, or a module.
    The string will be later composed with the configured message-prefix to build the exact topic name.

    This is the main logic that builds keys strings for topics/streaming, adding wildcards when needed

    Examples:
        build_routing_key(mymessaginglib.vision.control) -> "vision.control.#" routing with binding key
        build_routing_key(mymessaginglib.vision.control.Start) -> "vision.control.Start" direct routing
        build_routing_key(mymessaginglib.vision.control.Start()) -> "vision.control.Start" direct routing

    Args:
        pkg_or_msg: a Message instance, class or module to mymessaginglib codegen packages

    Returns: a routing key based on the type of message or package

    """
    backend = default_configuration.backend_config
    wildcard = backend.multi_wildcard_delimiter
    module_name = ""
    class_name = ""

    if isinstance(pkg_or_msg, betterproto.Message):
        module_name = pkg_or_msg.__module__
        class_name = pkg_or_msg.__class__.__name__
    elif isinstance(pkg_or_msg, type(betterproto.Message)):
        module_name = pkg_or_msg.__module__
        class_name = pkg_or_msg.__name__
    elif isinstance(pkg_or_msg, ModuleType):
        module_name = pkg_or_msg.__name__
        class_name = wildcard  # wildcard routing key for subscribing to all messages in a module

    return _build_routing_key(module_name, class_name)



from unittest.mock import ANY

import betterproto
import pytest

from . import tests

is_numpy_available = True
try:
    import numpy as np
    import numpy.testing as npt
except (ModuleNotFoundError, ImportError):
    is_numpy_available = False


@pytest.mark.skipif(not is_numpy_available, reason="numpy is not installed")
def test_complex_message() -> None:
    data = {
        "distance_cdf": np.array(
            [0.0, 0.1761976826283381, 0.3360906092222883, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0]
        ),
        "distance_cdf_step": np.float_(0.05),
        "part_bbox_extent": np.array([198.5670928955078, 201.40457153320312, 152.95013427734375]),
        "part_bbox_center": np.array([6.441764831542969, -19.629409790039062, 76.47506713867188]),
        "part_bbox_rotation": np.array(
            [
                [0.9950731992721558, -0.09914305806159973, 0.0],
                [0.09914308041334152, 0.9950731992721558, -0.0],
                [0.0, 0.0, 1.0],
            ]
        ),
        "scan_bbox_extent": np.array([198.28387451171875, 200.86383056640625, 153.09780883789062]),
        "scan_bbox_center": np.array([6.441764831542969, -19.629409790039062, 76.54890441894531]),
        "scan_bbox_rotation": np.array(
            [
                [0.9950731992721558, -0.09914305806159973, 0.0],
                [0.09914308041334152, 0.9950731992721558, -0.0],
                [0.0, 0.0, 1.0],
            ]
        ),
        "max_missing_area_centroid": np.array(
            [
                82.57372283935547,
                -29.880002975463867,
                15.93416690826416,
            ]
        ),
        "max_missing_area_size": np.int_(572),
        "part_pose": np.array(
            [
                [
                    -0.6797157804915129,
                    -0.34871117468390145,
                    0.6452803650368037,
                    55.72332582210536,
                ],
                [
                    -0.7334754848378484,
                    0.3235935191340345,
                    -0.5977465618459061,
                    -73.10345511815211,
                ],
                [
                    -0.0003676277833084182,
                    -0.8795952145834851,
                    -0.4757227565212457,
                    -0.24029829148940943,
                ],
                [1.0741441069939128e-10, 5.733221675185973e-10, -5.05052055732591e-10, 1.0],
            ]
        ),
    }

    # Build the Evaluation message
    msg = tests.TestMessage(
        content="test",
        number=12,
        detail="ed1b0017-1c49-4279-b3f3-c23195a4ed33",
        # `options` is a pb.commons.JsonContent instance but it's serialized internally by the library
        options=data,
    )
    # Serialize the Evaluation message. This will also serialize dictionaries
    # for JsonContent fields in the message.
    # Internally the dict is converted to a JsonContent value
    serialized = bytes(msg)
    # Deserialize (it will also deserialize the JsonContent to a dictionary
    # Note: numpy arrays are not preserved and are serialized back as lists.
    deserialized = tests.TestMessage().parse(serialized)
    # We assert that deserialized.options == data
    # Since the original data has numpy arrays as values, we use np.assert_array_equal
    for k, v in deserialized.options.items():
        npt.assert_array_equal(v, data[k])
    # Test to_dict and to_pydict
    res = msg.to_dict(casing=betterproto.Casing.SNAKE)
    for k, v in res["options"].items():
        npt.assert_array_equal(v, data[k])
    assert msg.to_dict(casing=betterproto.Casing.SNAKE) == dict(
        content="test",
        number="12",
        detail="ed1b0017-1c49-4279-b3f3-c23195a4ed33",
        options=ANY,
    )
    logged = deserialized.to_pydict(casing=betterproto.Casing.SNAKE)
    for k, v in logged["options"].items():
        npt.assert_array_equal(v, data[k])


def test_int_floats() -> None:
    msg = tests.tasks.TaskMessage(
        content="test",
        bbox=[1, -1, 1, 100],
        weights=[123.0, 0.5, 100_000],
    )
    # serialize
    ser = bytes(msg)
    # deserialize
    deser = tests.tasks.TaskMessage().parse(ser)
    assert msg.bbox == [round(v, 2) for v in deser.bbox]



"""Implements a Mosquitto (MQTT) Connection with both sync and async support"""
import asyncio
import logging
import os
import typing as tp
import uuid
from concurrent.futures import ThreadPoolExecutor

import aiomqtt
import can_ada

from ....config import default_configuration
from ....exceptions import ConnectionError, PublishError, RequeueMessage
from ....models import Envelope, IncomingMessageProtocol
from .. import BaseAsyncConnection

log = logging.getLogger(__name__)


VHOST = os.environ.get("MOSQUITTO_VHOST", "/")


async def connect() -> "Connection":
    """Get the singleton async connection."""
    conn = await Connection.get_connection(vhost=VHOST)
    return conn


async def reset_connection() -> "Connection":
    """Reset the singleton connection."""
    connection = await connect()
    await connection.disconnect()
    return await connect()


async def disconnect() -> None:
    connection = await connect()
    await connection.disconnect()


class Connection(BaseAsyncConnection):
    """Async Mosquitto Connection wrapper using aiomqtt."""

    instance_by_vhost: dict[str, "Connection | None"] = {}

    def __init__(
        self,
        username: str | None = None,
        password: str | None = None,
        host: str | None = None,
        port: int | None = None,
        url: str | None = None,
        worker_threads: int = 2,
        requeue_delay: int = 3,
        **kwargs,
    ):
        super().__init__()
        if url:
            # Parse the URL instead
            parsed = can_ada.parse(url)
            host = parsed.hostname or host
            port = parsed.port or port or 1883
            username = parsed.username
            password = parsed.password
        else:
            host or os.environ.get("MQTT_HOST") or os.environ.get("MOSQUITTO_HOST")
            port or os.environ.get("MQTT_PORT") or os.environ.get("MOSQUITTO_PORT")
            username or os.environ.get("MQTT_USERNAME") or os.environ.get("MOSQUITTO_USERNAME")
            password or os.environ.get("MQTT_PASSWORD") or os.environ.get("MOSQUITTO_PASSWORD")

        self.host = host or "127.0.0.1"
        self.port = int(port) if port else 1883
        self.username = username
        self.password = password

        self.requeue_delay = requeue_delay

        self.executor = ThreadPoolExecutor(
            max_workers=worker_threads
        )  # executor for sync callbacks
        self.consumers: dict[str, asyncio.Task] = {}
        self.queues: dict[str, dict] = {}
        self.stop_events: dict[str, asyncio.Event] = {}

        self._connection: aiomqtt.Client | None = None
        self._instance_lock: asyncio.Lock | None = None
        self._delimiter = default_configuration.backend_config.topic_delimiter  # e.g. "/"
        self._exchange = default_configuration.backend_config.namespace  # used as root prefix

    @property
    def is_connected_event(self) -> asyncio.Event:
        """Lazily create the event in the current running loop."""
        if self._is_connected_event is None:
            self._is_connected_event = asyncio.Event()
        return self._is_connected_event

    @property
    def lock(self) -> asyncio.Lock:
        if self._instance_lock is None:
            self._instance_lock = asyncio.Lock()
        return self._instance_lock

    def build_topic_key(self, topic: str) -> str:
        """Joins project prefix with topic using the configured delimiter."""
        return f"{self._exchange}{self._delimiter}{topic}"

    async def connect(self, timeout: float = 30.0) -> "Connection":
        async with self.lock:
            if self.is_connected():
                return self

            try:
                log.info("Connecting to Mosquitto at %s:%s", self.host, self.port)
                self._connection = aiomqtt.Client(
                    hostname=self.host,
                    port=self.port,
                    username=self.username,
                    password=self.password,
                    timeout=timeout,
                )
                # We enter the context manager to establish the connection
                await self._connection.__aenter__()
                self.is_connected_event.set()
                log.info("Successfully connected to Mosquitto")
                return self
            except Exception as e:
                self.is_connected_event.clear()
                self._connection = None
                log.exception("Failed to connect to Mosquitto")
                raise ConnectionError(f"MQTT Connect failed: {e}")

    async def disconnect(self, **kwargs) -> None:
        async with self.lock:
            for tag, task in self.consumers.items():
                task.cancel()
            # Wait for them to finish so they exit their 'async with client' blocks
            if self.consumers:
                await asyncio.gather(*self.consumers.values(), return_exceptions=True)
            try:
                if self._connection:
                    await self._connection.__aexit__(None, None, None)
                    self._connection = None
            except Exception as e:
                log.warning("Failed to disconnect from Mosquitto: %s", e)

            self.is_connected_event.clear()
            log.info("Mosquitto connection closed")

    async def publish(self, topic: str, message: IncomingMessageProtocol, **kwargs) -> None:
        if not self.is_connected():
            raise ConnectionError("Not connected")

        topic_key = self.build_topic_key(topic)
        log.debug("MQTT Publish to %s", topic_key)

        # MQTT handles the distribution. If it's a task, the subscribers
        # will handle the 'shared' logic via their subscription string.
        await self._connection.publish(
            topic_key,
            payload=message.body,
            qos=kwargs.get("qos", 1),
            retain=kwargs.get("retain", False),
        )

    async def subscribe(self, topic: str, callback: tp.Callable, shared: bool = False) -> str:
        """Subscribes to a topic and starts a consumer loop in the background.
        Args:
            topic
            callback
            shared

        Returns:
            The consumer tag
        """
        timeout = 3.0
        log.debug("Subscribing topic: %s", topic)
        async with self.lock:
            queue_meta = await self.setup_queue(topic, shared)
            tag = queue_meta["tag"]
            log.debug("Subscribing consumer: %s", tag)
            ready_event = asyncio.Event()
            stop_event = asyncio.Event()
            task = asyncio.create_task(
                self._consumer_loop(queue_meta["sub_key"], callback, ready_event, stop_event)
            )

            try:
                await asyncio.wait_for(ready_event.wait(), timeout=timeout)
                self.consumers[tag] = task
                self.stop_events[tag] = stop_event
                log.debug("Subscribed consumer: %s", tag)
                return tag
            except asyncio.TimeoutError:
                stop_event.set()
                task.cancel()
                raise TimeoutError(f"Failed to subscribe to {topic} within {timeout}s.")

    async def _consumer_loop(
        self,
        sub_key: str,
        callback: tp.Callable,
        ready_event: asyncio.Event,
        stop_event: asyncio.Event,
    ):
        # Instantiate a NEW client locally instead of reusing self._connection
        client = aiomqtt.Client(
            hostname=self.host,
            port=self.port,
            username=self.username,
            password=self.password,
        )
        try:
            async with client:
                await client.subscribe(sub_key)
                ready_event.set()
                # In 3.10, we manually iterate the generator to apply timeouts
                messages_iter = client.messages.__aiter__()
                # The loop needs to handle the stop_event signal even if no messages arrive
                while not stop_event.is_set():
                    try:
                        # Python 3.10 compatible timeout
                        message = await asyncio.wait_for(messages_iter.__anext__(), timeout=1.0)
                        # wrap the iterator in a time out context manager to avoid blocking if no messages arrive
                        if stop_event.is_set():
                            break
                        original_topic = message.topic.value.removeprefix(
                            f"{self._exchange}{self._delimiter}"
                        )
                        envelope = Envelope(body=message.payload, routing_key=original_topic)
                        await self._handle_callback(callback, envelope, original_topic)
                    except asyncio.TimeoutError:
                        continue  # Just a heartbeat to check stop_event

        except asyncio.CancelledError:
            pass
        finally:
            log.debug("Consumer loop stopped for %s", sub_key)

    async def _handle_callback(self, callback: tp.Callable, envelope: Envelope, topic: str):
        """Executes callback with Requeue logic similar to Redis/Rabbit."""
        try:
            if asyncio.iscoroutinefunction(callback):
                await callback(envelope)
            else:
                await asyncio.get_event_loop().run_in_executor(self.executor, callback, envelope)
        except RequeueMessage:
            log.warning("Requeuing MQTT message for %s", topic)
            await asyncio.sleep(self.requeue_delay)
            await self.publish(topic, envelope)
        except Exception as exc:
            log.exception("Callback failed for topic %s", topic)
            raise PublishError(f"Failed to publish message to topic {topic}") from exc

    async def unsubscribe(self, tag: str, **kwargs):
        """Gracefully stops a consumer loop and unsubscribes from the broker."""
        async with self.lock:
            queue = self.queues.get(tag, None)
            log.debug("Unsubscribing tag: %s - %s", tag, str(queue) or "No queue found")
            stop_event = self.stop_events.get(tag)
            task = self.consumers.get(tag)

            if stop_event:
                stop_event.set()  # Signals the _consumer_loop to break

            if task:
                # We wait for the consumer loop to exit naturally and perform its cleanup
                try:
                    await asyncio.wait_for(task, timeout=5.0)
                except asyncio.TimeoutError:
                    task.cancel()  # Force kill if it doesn't stop
            # Cleanup state
            _ = self.consumers.pop(tag, None)
            self.stop_events.pop(tag, None)
            queue = self.queues.pop(tag, None)
            log.info("Successfully unsubscribed tag: %s", tag)

    async def purge(self, topic: str, **kwargs) -> None:
        """
        Clears the retained message for the topic.
        Note: This does not affect messages currently in flight to
        offline clients with persistent sessions.
        """

        async with self.lock:
            if not self.is_connected():
                raise ConnectionError("Not connected to Mosquitto")

            topic_key = self.build_topic_key(topic)
            sub_key = f"$share/shared_group/{topic_key}"
            log.debug("Purging shared topic: %s", sub_key)
            # Publishing a zero-length payload with retain=True deletes the retained message
            await self._connection.publish(sub_key, payload=None, qos=1, retain=True)

    def get_message_count(self, topic: str) -> int:
        # Mosquitto doesn't have a count messages API
        raise NotImplementedError

    def get_consumer_count(self, topic: str) -> int:
        # TODO
        raise NotImplementedError

    async def setup_queue(self, topic: str, shared: bool = False) -> dict:
        tag = f"consumer_{uuid.uuid4().hex[:8]}"
        topic_key = self.build_topic_key(topic)  # as used in Mosquitto
        group_name = "shared_group" if shared else ""

        # MQTT 5.0 Shared Subscription: $share/<group_name>/<topic>
        if shared:
            sub_key = f"$share/shared_group/{topic_key}"
        else:
            sub_key = topic_key

        queue = {
            "is_shared": shared,
            "group_name": group_name,
            "sub_key": sub_key,
            "tag": tag,
            "topic": topic,
            "topic_key": topic_key,
        }
        self.queues[tag] = queue
        return queue



class RequeueMessage(Exception):
    """Raise when a message could not be handled but should be requeued."""

    ...


class ConnectionError(Exception):
    """Raised when connection operations fail."""

    ...


class PublishError(Exception):
    """Raised when a callback fails to process a message and doesn't raise the RequeueMessage exception."""

    ...



import asyncio
import functools
import logging
import threading
import typing as tp
from abc import ABC, abstractmethod

from ..exceptions import RequeueMessage
from ..helpers import get_backend
from ..models import (
    BaseQueue,
    IncomingMessageProtocol,
    LoggerCallback,
    ProtoBunnyMessage,
    SyncCallback,
    default_configuration,
    deserialize_message,
    deserialize_result_message,
    get_body,
)

log = logging.getLogger(__name__)


class BaseConnection(ABC):
    username: str | None
    password: str | None
    host: str | None
    port: int | None
    vhost: str
    worker_threads: int
    prefetch_count: int
    requeue_delay: float
    exchange_name: str | None
    dl_exchange: str | None
    dl_queue: str | None
    heartbeat: int | None
    timeout: int | None
    url: str | None = None
    queues: dict[str, tp.Any] = {}

    @abstractmethod
    def __init__(self, *args, **kwargs):
        ...

    @abstractmethod
    def publish(
        self,
        topic: str,
        message: "IncomingMessageProtocol",
        **kwargs,
    ) -> None | tp.Awaitable[None]:
        ...

    @abstractmethod
    def disconnect(self, timeout: float = 30) -> None | tp.Awaitable[None]:
        ...

    @classmethod
    @abstractmethod
    def get_connection(cls, vhost: str = "") -> tp.Any | tp.Awaitable[tp.Any]:
        ...

    @abstractmethod
    def is_connected(self) -> bool | tp.Awaitable[bool]:
        ...

    @abstractmethod
    def connect(self, timeout: float = 30) -> None | tp.Awaitable[None]:
        ...

    @abstractmethod
    def subscribe(
        self, topic: str, callback: SyncCallback, shared: bool = False
    ) -> str | tp.Awaitable[str]:
        ...

    @abstractmethod
    def unsubscribe(self, topic: str, **kwargs) -> None | tp.Awaitable[None]:
        ...

    @abstractmethod
    def purge(self, topic: str, **kwargs) -> None | tp.Awaitable[None]:
        ...

    @abstractmethod
    def get_message_count(self, topic: str) -> int | tp.Awaitable[int]:
        ...

    @abstractmethod
    def get_consumer_count(self, topic: str) -> int | tp.Awaitable[int]:
        ...

    @abstractmethod
    def setup_queue(self, topic: str, shared: bool) -> tp.Any | tp.Awaitable[tp.Any]:
        ...


class BaseAsyncConnection(BaseConnection, ABC):
    instance_by_vhost: dict[str, "BaseAsyncConnection"]

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.vhost = kwargs.get("vhost", "")


class BaseSyncConnection(BaseConnection, ABC):
    _lock: threading.RLock
    _stopped: asyncio.Event | None
    instance_by_vhost: dict[str, "BaseSyncConnection"]
    async_class: "type[BaseAsyncConnection] | None"

    def __init__(self, **kwargs):
        """Initialize sync connection.

        It's a wrapper around the async connection
        The Async connection is in _async_conn attribute

        Args:
            **kwargs: Same arguments as AsyncConnection
        """
        super().__init__()
        self._async_conn = self.get_async_connection(**kwargs)
        self._thread: threading.Thread | None = None
        self._loop: asyncio.AbstractEventLoop | None = None
        self._ready = threading.Event()
        self._stopped: asyncio.Event | None = None
        self.vhost = self._async_conn.vhost
        self._started = False
        self.instance_by_vhost = {}

    def get_async_connection(self, **kwargs) -> "BaseAsyncConnection":
        if hasattr(self, "_async_conn"):
            return self._async_conn
        return self.async_class(**kwargs)

    def _run_loop(self) -> None:
        """Run the event loop in a dedicated thread."""
        loop = None
        try:
            # Create a fresh loop for this specific thread
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)
            self._loop = loop

            # Run the 'stop' watcher
            loop.create_task(self._async_run_watcher())

            # Signal readiness NOW that self._loop is assigned and running
            self._ready.set()
            loop.run_forever()
        except asyncio.CancelledError:
            pass
        except Exception:
            log.exception("Event loop thread crashed")
        finally:
            if loop:
                loop.close()
            self._loop = None
            log.info("Event loop thread stopped")

    async def _async_run_watcher(self) -> None:
        """Wait for the stop signal inside the loop."""
        self._stopped = asyncio.Event()
        await self._stopped.wait()
        asyncio.get_running_loop().stop()

    async def _async_run(self) -> None:
        """Async event loop runner."""
        self._loop = asyncio.get_running_loop()
        self._stopped = asyncio.Event()
        self._loop.call_soon_threadsafe(self._ready.set)
        await self._stopped.wait()

    def _ensure_loop(self) -> None:
        """Ensure event loop thread is running.

        Raises:
            ConnectionError: If event loop fails to start
        """
        # Check if the thread exists AND is actually running
        if self._thread and self._thread.is_alive() and self._loop and self._loop.is_running():
            return

        log.info("Starting (or restarting) wrapping event loop thread")

        # Reset state for a fresh start
        self._ready.clear()
        self._loop = None

        self._thread = threading.Thread(
            target=self._run_loop, name="protobunny_event_loop", daemon=True
        )
        self._thread.start()

        if not self._ready.wait(timeout=10.0):
            # Cleanup on failure to prevent stale state for next attempt
            self._thread = None
            raise ConnectionError("Event loop thread failed to start or signal readiness")

    def _run_coro(self, coro, timeout: float | None = None):
        """Run a coroutine in the event loop thread and return result.

        Args:
            coro: The coroutine to run
            timeout: Maximum time to wait for result (seconds)

        Returns:
            The coroutine result

        Raises:
            TimeoutError: If operation times out
            ConnectionError: If event loop is not available
        """
        self._ensure_loop()
        if self._loop is None:
            raise ConnectionError("Event loop not initialized")

        future = asyncio.run_coroutine_threadsafe(coro, self._loop)
        try:
            return future.result(timeout=timeout)
        except TimeoutError:
            future.cancel()
            raise

    def is_connected(self) -> bool:
        """Check if connection is established."""
        if not self._loop or not self._loop.is_running():
            return False
        return self._async_conn.is_connected()

    @classmethod
    def get_connection(cls, vhost: str = "") -> "BaseSyncConnection":
        """Get singleton instance (sync)."""
        with cls._lock:
            if not cls.instance_by_vhost.get(vhost):
                cls.instance_by_vhost[vhost] = cls(vhost=vhost)
            if not cls.instance_by_vhost[vhost].is_connected():
                cls.instance_by_vhost[vhost].connect()
            return cls.instance_by_vhost[vhost]

    def publish(
        self,
        topic: str,
        message: "IncomingMessageProtocol",
        mandatory: bool = False,
        immediate: bool = False,
        timeout: float = 10.0,
    ) -> None:
        """Publish a message to a topic.

        Args:
            topic: The routing key/topic
            message: The message to publish
            mandatory: If True, raise error if message cannot be routed
            immediate: If True, publish message immediately to the queue
            timeout: Maximum time to wait for publish (seconds)

        Raises:
            ConnectionError: If not connected
            TimeoutError: If operation times out
        """
        self._run_coro(
            self._async_conn.publish(topic, message, mandatory=mandatory, immediate=immediate),
            timeout=timeout,
        )

    def setup_queue(self, topic: str, shared: bool, timeout: int = 10) -> tp.Any:
        self._run_coro(self._async_conn.setup_queue(topic, shared), timeout=timeout)

    def subscribe(
        self, topic: str, callback: tp.Callable, shared: bool = False, timeout: float = 10.0
    ) -> str:
        """Subscribe to a queue/topic.

        Args:
            topic: The routing key/topic to subscribe to
            callback: Function to handle incoming messages
            shared: if True, use shared queue (round-robin delivery)
            timeout: Maximum time to wait for subscription (seconds)

        Returns:
            Subscription tag identifier

        Raises:
            ConnectionError: If not connected
            TimeoutError: If operation times out
        """
        return self._run_coro(self._async_conn.subscribe(topic, callback, shared), timeout=timeout)

    def unsubscribe(
        self, tag: str, timeout: float = 10.0, if_unused: bool = True, if_empty: bool = True
    ) -> None:
        """Unsubscribe from a queue.

        Args:
            if_unused:
            if_empty:
            tag: Subscription identifier returned from subscribe()
            timeout: Maximum time to wait (seconds)

        Raises:
            TimeoutError: If operation times out
        """
        self._run_coro(
            self._async_conn.unsubscribe(tag, if_empty=if_empty, if_unused=if_unused),
            timeout=timeout,
        )

    def purge(self, topic: str, timeout: float = 10.0, **kwargs) -> None:
        """Empty a queue of all messages.

        Args:
            topic: The queue topic to purge
            timeout: Maximum time to wait (seconds)

        Raises:
            ConnectionError: If not connected
            TimeoutError: If operation times out
        """
        self._run_coro(self._async_conn.purge(topic, **kwargs), timeout=timeout)

    def get_message_count(self, topic: str, timeout: float = 10.0) -> int:
        """Get the number of messages in a queue.

        Args:
            topic: The queue topic
            timeout: Maximum time to wait (seconds)

        Returns:
            Number of messages in the queue

        Raises:
            ConnectionError: If not connected
            TimeoutError: If operation times out
        """
        return self._run_coro(self._async_conn.get_message_count(topic), timeout=timeout)

    def get_consumer_count(self, topic: str, timeout: float = 10.0) -> int:
        """Get the number of messages in a queue.

        Args:
            topic: The queue topic
            timeout: Maximum time to wait (seconds)

        Returns:
            Number of messages in the queue

        Raises:
            ConnectionError: If not connected
            TimeoutError: If operation times out
        """
        return self._run_coro(self._async_conn.get_consumer_count(topic), timeout=timeout)

    def __enter__(self):
        """Context manager entry."""
        self.connect()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Context manager exit."""
        self.disconnect()
        return False

    def connect(self, timeout: float = 10.0) -> None:
        """Establish Sync connection.

        Args:
            timeout: Maximum time to wait for connection (seconds)

        Raises:
            ConnectionError: If connection fails
            TimeoutError: If connection times out
        """
        self._run_coro(self._async_conn.connect(timeout), timeout=timeout)
        self.__class__.instance_by_vhost[self.vhost] = self

    def disconnect(self, timeout: float = 10.0) -> None:
        """Close sync and the underlying async connections and stop event loop.

        Args:
            timeout: Maximum time to wait for cleanup (seconds)
        """
        with self._lock:
            try:
                if self._loop and self._loop.is_running():
                    self._run_coro(self._async_conn.disconnect(timeout), timeout=timeout)
                # Stop the loop (see _async_run_watcher)
                if self._stopped and self._loop:
                    self._loop.call_soon_threadsafe(self._stopped.set)
            except Exception as e:
                log.warning("Async disconnect failed during sync shutdown: %s", e)
            finally:
                if self._thread and self._thread.is_alive():
                    self._thread.join(timeout=5.0)
                    if self._thread.is_alive():
                        log.warning("Event loop thread did not stop within timeout")
                self._started = None
                self._loop = None
                self._thread = None
                self.async_class.instance_by_vhost.pop(self.vhost, None)
            type(self).instance_by_vhost.pop(self.vhost, None)


class BaseSyncQueue(BaseQueue, ABC):
    def get_connection(self) -> BaseConnection:
        backend = get_backend()
        return backend.connection.connect()

    def publish(self, message: "ProtoBunnyMessage") -> None:
        """Publish a message to the queue.

        Args:
            message: a ProtoBunnyMessage message
        """
        self.send_message(self.topic, bytes(message))

    def publish_result(
        self,
        result: "Result",
        topic: str | None = None,
        correlation_id: str | None = None,
    ) -> None:
        """Publish a message to the results topic.

        Args:
            result: a amlogic_messages.results.Result message
            topic:
            correlation_id:
        """
        result_topic = topic or self.result_topic
        log.info("Publishing result to: %s", result_topic)
        self.send_message(
            result_topic, bytes(result), correlation_id=correlation_id, persistent=False
        )

    def _receive(
        self, callback: "SyncCallback | LoggerCallback", message: "IncomingMessageProtocol"
    ) -> None:
        """Handle a message from the queue.

        Args:
            callback: a callable accepting a message as only argument.
            message: the aio_pika.IncomingMessage object received from the queue.
        """
        if not message.routing_key:
            raise ValueError("Routing key was not set. Invalid topic")
        delimiter = default_configuration.backend_config.topic_delimiter
        if message.routing_key == self.result_topic or message.routing_key.endswith(
            f"{delimiter}result"
        ):
            # Skip a result message. Handling result messages happens in `_receive_results` method.
            # In case the subscription has .# as binding key,
            # this method catches also results message for all the topics in that namespace.
            return
        # msg: "ProtoBunnyMessage" = deserialize_message(message.routing_key, message.body)
        try:
            callback(deserialize_message(message.routing_key, message.body))
        except RequeueMessage:
            raise
        except Exception as exc:  # pylint: disable=W0703
            log.exception("Could not process message: %s", str(message.body))
            msg = deserialize_message(message.routing_key, message.body)
            if not msg:
                log.warning(
                    "Could not deserialize with routing key: %s - not publishing a result",
                    str(message.routing_key),
                )
                return
            result = msg.make_result(return_code=ReturnCode.FAILURE, error=str(exc))
            self.publish_result(
                result, topic=msg.result_topic, correlation_id=message.correlation_id
            )

    def subscribe(self, callback: "SyncCallback | LoggerCallback") -> None:
        """Subscribe to messages from the queue.

        Args:
            callback:

        """
        if self.subscription is not None:
            log.warning("Already subscribed...")
            return
        func = functools.partial(self._receive, callback)
        self.subscription = self.get_connection().subscribe(
            self.topic, func, shared=self.shared_queue
        )

    def unsubscribe(self, if_unused: bool = True, if_empty: bool = True) -> None:
        """Unsubscribe from the queue."""
        if self.subscription is not None:
            self.get_connection().unsubscribe(
                self.get_tag(), if_unused=if_unused, if_empty=if_empty
            )
            self.subscription = None

    def _receive_result(
        self,
        callback: tp.Callable[["Result"], tp.Any],
        message: IncomingMessageProtocol,
    ) -> None:
        """Handle a message from the queue.

        Args:
            callback : function to call with deserialized result.
                Accept parameters like (message: Message, return_code: int, return_value: dict, error:str)
            message : an `aio_pika.IncomingMessage` or Envelope serialized message from the queue.
        """
        try:
            result = deserialize_result_message(message.body)
            """
            `result.source_message` is a protobuf.Any instance.
            It has `type_url` property that describes the type of message.
            To reconstruct the source message you can  do it by using the Result.source property or
            base methods.
            >>> source_message = result.source
            or more explicitly
            >> message_type = get_message_class_from_type_url(result.source_message.type_url)
            >> source_message = message_type().parse(result.source_message.value)
            """
            callback(result)
        except Exception:  # pylint: disable=W0703
            log.exception("Could not process result: %s", str(message.body))

    def subscribe_results(self, callback: tp.Callable[["Result"], tp.Any]) -> None:
        """Subscribe to results from the queue.

        See the deserialize_result method for return params.

        Args:
            callback : function to call when results come in.
        """
        if self.subscription is not None:
            log.warning("Already subscribed...")
            return
        func = functools.partial(self._receive_result, callback)
        self.result_subscription = self.get_connection().subscribe(
            self.result_topic, func, shared=False
        )

    def unsubscribe_results(self) -> None:
        """Unsubscribe from results. Will always delete the underlying queues"""
        if self.result_subscription is not None:
            self.get_connection().unsubscribe(
                self.result_subscription, if_unused=False, if_empty=False
            )
            self.result_subscription = None

    def purge(self, **kwargs) -> None:
        """Delete all messages from the queue."""
        if not self.shared_queue:
            raise RuntimeError("Can only purge shared queues")
        self.get_connection().purge(self.topic, **kwargs)

    def get_message_count(self) -> int:
        """Get current message count."""
        if not self.shared_queue:
            raise RuntimeError("Can only get count of shared queues")
        log.debug("Getting queue message count")
        return self.get_connection().get_message_count(self.topic)

    def get_consumer_count(self) -> int:
        """Get current message count."""
        if not self.shared_queue:
            raise RuntimeError("Can only get count of shared queues")
        log.debug("Getting queue message count")
        return self.get_connection().get_consumer_count(self.topic)


class LoggingSyncQueue(BaseSyncQueue):
    """Represents a specialized queue for logging purposes.

    >>> import protobunny as pb
    >>> pb.subscribe_logger()  # it uses the default logger_callback

    You can add a custom callback that accepts the envelope message from the backend and msg_content: str as arguments.
    The type of message will respect the protocol IncomingMessageProtocol

    >>> def log_callback(message: "IncomingMessageProtocol", msg_content: str):
    >>>     print(message.body)
    >>> pb.subscribe_logger(log_callback)

    You can use functools.partial to add more arguments

    >>> def log_callback_with_args(message: aio_pika.IncomingMessage, msg_content: str, maxlength: int):
    >>>     print(message.body[maxlength])
    >>> import functools
    >>> functools.partial(log_callback_with_args, maxlength=100)
    >>> pb.subscribe_logger(log_callback_with_args)
    """

    async def send_message(
        self,
        topic: str,
        content: bytes,
        correlation_id: str | None = None,
        persistent: bool = False,
    ) -> None:
        # This queue is only for receiving messages, so it doesn't need to send messages'
        raise NotImplementedError()

    def __init__(self, prefix: str) -> None:
        backend = default_configuration.backend_config
        delimiter = backend.topic_delimiter
        wildcard = backend.multi_wildcard_delimiter
        prefix = prefix or default_configuration.messages_prefix
        super().__init__(f"{prefix}{delimiter}{wildcard}")

    @property
    def result_topic(self) -> str:
        return ""

    def publish(self, message: "ProtoBunnyMessage") -> None:
        raise NotImplementedError

    def publish_result(
        self,
        result: "Result",
        topic: str | None = None,
        correlation_id: str | None = None,
    ) -> None:
        raise NotImplementedError

    def _receive(
        self,
        log_callback: "LoggerCallback",
        message: IncomingMessageProtocol,
    ):
        """Call the logging callback.

        Args:
            log_callback: The callback function passed to pb.subscribe_logger().
              It receives the aio_pika IncomingMessage as first argument and the string to log as second.

            message: the IncomingMessage
        """
        if message.routing_key is None:
            raise ValueError("Routing key was not set. Invalid topic")
        try:
            body = get_body(message)
            log_callback(message, body)
        except RequeueMessage:
            raise
        except Exception as exc:  # pylint: disable=W0703
            log.exception(
                "Could not process message on Logging queue: %s - %s", str(message.body), str(exc)
            )

    def get_tag(self) -> str:
        return self.topic


def is_task(topic: str) -> bool:
    delimiter = default_configuration.backend_config.topic_delimiter
    return "tasks" in topic.split(delimiter)


# keep always the imports of generated code at the end of the file
from ..core.results import Result, ReturnCode



import asyncio
import functools
import logging
import typing as tp
from abc import ABC, abstractmethod

from ...exceptions import RequeueMessage
from ...helpers import get_backend
from ...models import (
    AsyncCallback,
    BaseQueue,
    Envelope,
    IncomingMessageProtocol,
    LoggerCallback,
    ProtoBunnyMessage,
    SyncCallback,
    default_configuration,
    deserialize_message,
    deserialize_result_message,
    get_body,
)

if tp.TYPE_CHECKING:
    # This only runs during IDE linting or Mypy runs
    from typing_extensions import Self
else:
    from typing import Any as Self

log = logging.getLogger(__name__)


class BaseConnection(ABC):
    username: str | None
    password: str | None
    host: str | None
    port: int | None
    vhost: str
    worker_threads: int
    prefetch_count: int
    requeue_delay: float
    exchange_name: str | None
    dl_exchange: str | None
    dl_queue: str | None
    heartbeat: int | None
    timeout: int | None
    url: str | None = None
    queues: dict[str, tp.Any] = {}

    @abstractmethod
    def __init__(self, *args, **kwargs):
        ...

    @abstractmethod
    def publish(
        self,
        topic: str,
        message: "IncomingMessageProtocol",
        **kwargs,
    ) -> None | tp.Awaitable[None]:
        ...

    @abstractmethod
    def disconnect(self, timeout: float = 30) -> None | tp.Awaitable[None]:
        ...

    @classmethod
    @abstractmethod
    def get_connection(cls, vhost: str = "") -> tp.Any | tp.Awaitable[tp.Any]:
        ...

    @abstractmethod
    def is_connected(self) -> bool:
        ...

    @abstractmethod
    def connect(self, timeout: float = 30) -> None | tp.Awaitable[None]:
        ...

    @abstractmethod
    def subscribe(
        self, topic: str, callback: SyncCallback, shared: bool = False
    ) -> str | tp.Awaitable[str]:
        ...

    @abstractmethod
    def unsubscribe(self, topic: str, **kwargs) -> None | tp.Awaitable[None]:
        ...

    @abstractmethod
    def purge(self, topic: str, **kwargs) -> None | tp.Awaitable[None]:
        ...

    @abstractmethod
    def get_message_count(self, topic: str) -> int | tp.Awaitable[int]:
        ...

    @abstractmethod
    def get_consumer_count(self, topic: str) -> int | tp.Awaitable[int]:
        ...

    @abstractmethod
    def setup_queue(self, topic: str, shared: bool) -> tp.Any | tp.Awaitable[tp.Any]:
        ...


class BaseAsyncConnection(BaseConnection, ABC):
    _lock: asyncio.Lock | None = None
    instance_by_vhost: dict[str, Self] = {}

    @abstractmethod
    async def connect(self, timeout: float = 30) -> None:
        ...

    @abstractmethod
    async def disconnect(self, timeout: float = 30) -> None:
        ...

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._is_connected_event: asyncio.Event | None = None
        self.vhost = kwargs.get("vhost", "")
        self._loop: asyncio.AbstractEventLoop | None = None
        try:
            self._loop = asyncio.get_running_loop()
        except RuntimeError:
            # Fallback if __init__ is called outside a running loop
            # (though get_connection should be called inside one)
            self._loop = None

    @classmethod
    def _get_class_lock(cls) -> asyncio.Lock:
        """Ensure the class lock is bound to the current running loop."""
        if cls._lock is None:
            cls._lock = asyncio.Lock()
        return cls._lock

    @property
    @abstractmethod
    def is_connected_event(self) -> asyncio.Event:
        ...

    @classmethod
    async def get_connection(cls, vhost: str = "/") -> Self:
        """Get singleton instance (async)."""
        current_loop = asyncio.get_running_loop()
        async with cls._get_class_lock():
            instance = cls.instance_by_vhost.get(vhost)
            # Check if we have an instance AND if it belongs to the CURRENT loop
            if instance:
                # We need to check if the instance's internal loop matches our current loop
                # and if that loop is actually still running.
                if instance._loop != current_loop or not instance.is_connected_event.is_set():
                    log.warning("Found stale connection for %s (loop mismatch). Resetting.", vhost)
                    await instance.disconnect()  # Cleanup the old one
                    instance = None

            if instance is None:
                log.debug("Creating fresh connection for %s", vhost)
                new_instance = cls(vhost=vhost)
                new_instance._loop = current_loop  # Store the loop it was born in
                await new_instance.connect()
                cls.instance_by_vhost[vhost] = new_instance
                instance = new_instance
            return instance

    def is_connected(self) -> bool:
        """Check if connection is established and healthy."""
        return self.is_connected_event.is_set()


class BaseAsyncQueue(BaseQueue, ABC):
    async def get_connection(self) -> "BaseAsyncConnection":
        backend = get_backend()
        return await backend.connection.connect()

    async def publish(self, message: ProtoBunnyMessage) -> None:
        """Publish a message to the queue.

        Args:
            message: a protobuf message
        """
        await self.send_message(self.topic, bytes(message))

    async def _receive(
        self, callback: "AsyncCallback | LoggerCallback", message: IncomingMessageProtocol
    ) -> None:
        """Handle a message from the queue.

        Args:
            callback: a callable accepting a message as only argument.
            message: the IncomingMessageProtocol object received from the queue.
        """
        delimiter = default_configuration.backend_config.topic_delimiter
        if not message.routing_key:
            raise ValueError("Routing key was not set. Invalid topic")
        if message.routing_key == self.result_topic or message.routing_key.endswith(
            f"{delimiter}result"
        ):
            # Skip a result message. Handling result messages happens in `_receive_results` method.
            # In case the subscription has .# as binding key,
            # this method would catch results message for all the topics in that namespace.
            return

        msg: "ProtoBunnyMessage" = deserialize_message(message.routing_key, message.body)
        try:
            await callback(msg)
        except RequeueMessage:
            raise
        except Exception as exc:  # pylint: disable=W0703
            log.exception("Could not process message: %s", str(message.body))
            result = msg.make_result(return_code=ReturnCode.FAILURE, error=str(exc))
            await self.publish_result(
                result, topic=self.result_topic, correlation_id=message.correlation_id
            )

    async def subscribe(self, callback: "AsyncCallback | LoggerCallback") -> None:
        """Subscribe to messages from the queue.

        Args:
            callback: The user async callback to call when a message is received.
              The callback should accept a single argument of type `ProtoBunnyMessage`.

        Note: The real callback that consumes the incoming aio-pika message is the method AsyncConnection._on_message
        The AsyncQueue._receive method is called from there to deserialize the message and in turn calls the user callback.
        """
        if self.subscription is not None:
            log.warning("Already subscribed...")
            return

        func = functools.partial(self._receive, callback)
        conn = await self.get_connection()
        self.subscription = await conn.subscribe(self.topic, func, shared=self.shared_queue)

    async def unsubscribe(self, if_unused: bool = True, if_empty: bool = True) -> None:
        """Unsubscribe from the queue."""
        if self.subscription is not None:
            conn = await self.get_connection()
            await conn.unsubscribe(self.get_tag(), if_unused=if_unused, if_empty=if_empty)
            self.subscription = None

    async def publish_result(
        self,
        result: "Result",
        topic: str | None = None,
        correlation_id: str | None = None,
    ) -> None:
        """Publish a message to the results topic.

        Args:
            result: a amlogic_messages.results.Result message
            topic:
            correlation_id:
        """
        result_topic = topic or self.result_topic
        log.info("Publishing result to: %s", result_topic)
        await self.send_message(
            result_topic, bytes(result), correlation_id=correlation_id, persistent=False
        )

    async def _receive_result(
        self,
        callback: "AsyncCallback",
        message: IncomingMessageProtocol,
    ) -> None:
        """Handle a message from the queue.

        Args:
            callback : function to call with deserialized result.
                Accept parameters like (message: Message, return_code: int, return_value: dict, error:str)
            message : `IncomingMessageProtocol` serialized message from the queue.
        """
        try:
            result = deserialize_result_message(message.body)
            # `result.source_message` is a protobuf.Any instance.
            # It has `type_url` property that describes the type of message.
            # To reconstruct the source message you can  do it by using the Result.source property or
            # base methods.
            # >>> source_message = result.source
            # or more explicitly
            # >> message_type = get_message_class_from_type_url(result.source_message.type_url)
            # >> source_message = message_type().parse(result.source_message.value)
            await callback(result)
        except Exception:
            log.exception("Could not process result: %s", str(message.body))

    async def subscribe_results(self, callback: "AsyncCallback") -> None:
        """Subscribe to results from the queue.

        See the deserialize_result method for return params.

        Args:
            callback : function to call when results come in.
        """
        if self.subscription is not None:
            log.warning("Already subscribed...")
            return
        func = functools.partial(self._receive_result, callback)
        conn = await self.get_connection()
        self.result_subscription = await conn.subscribe(self.result_topic, func, shared=False)

    async def unsubscribe_results(self) -> None:
        """Unsubscribe from results. Will always delete the underlying queues"""
        if self.result_subscription is not None:
            conn = await self.get_connection()
            await conn.unsubscribe(self.result_subscription, if_unused=False, if_empty=False)
            self.result_subscription = None

    async def purge(self, reset_groups: bool = False) -> None:
        """Delete all messages from the queue."""
        if not self.shared_queue:
            raise RuntimeError("Can only purge shared queues")
        conn = await self.get_connection()
        await conn.purge(self.topic, reset_groups=reset_groups)

    async def get_message_count(self) -> int | None:
        """Get current message count."""
        if not self.shared_queue:
            raise RuntimeError("Can only get count of shared queues")
        conn = await self.get_connection()
        return await conn.get_message_count(self.topic)

    async def get_consumer_count(self) -> int | None:
        """Get current message count."""
        if not self.shared_queue:
            raise RuntimeError("Can only get count of shared queues")
        conn = await self.get_connection()
        return await conn.get_consumer_count(self.topic)

    def get_tag(self) -> str:
        return self.subscription

    @staticmethod
    async def send_message(
        topic: str, body: bytes, correlation_id: str | None = None, persistent: bool = True
    ) -> None:
        """Low-level message sending implementation.

        Args:
            topic: a topic name for direct routing or a routing key with special binding keys
            body: serialized message (e.g. a serialized protobuf message or a json string)
            correlation_id: is present for result messages
            persistent: if true will use aio_pika.DeliveryMode.PERSISTENT

        Returns:

        """
        backend = get_backend()
        message = Envelope(
            body=body,
            correlation_id=correlation_id or b"",
        )
        conn = await backend.connection.connect()
        await conn.publish(topic, message)


class LoggingAsyncQueue(BaseAsyncQueue):
    """Represents a specialized queue for logging purposes.

    >>> from protobunny import asyncio as pb
    >>> async def add_logger():
    >>>     await pb.subscribe_logger()  # it uses the default logger_callback

    You can add a custom callback that accepts message: aio_pika.IncomingMessage, msg_content: str as arguments.
    Note that the callback must be sync even for the async logger and
    it must be a function who purely calls the logging module and can perform other non IO operations

    >>> def log_callback(message, msg_content: str):
    >>>     print(message.body)
    >>> async def add_logger():
    >>>     await pb.subscribe_logger(log_callback)

    You can use functools.partial to add more arguments

    >>> def log_callback_with_args(message, msg_content: str, maxlength: int):
    >>>     print(message.body[maxlength])
    >>> import functools
    >>> functools.partial(log_callback_with_args, maxlength=100)
    >>> async def add_logger():
    >>>     await pb.subscribe_logger(log_callback_with_args)
    """

    def __init__(self, prefix: str) -> None:
        backend = default_configuration.backend_config
        delimiter = backend.topic_delimiter
        wildcard = backend.multi_wildcard_delimiter
        prefix = prefix or default_configuration.messages_prefix
        super().__init__(f"{prefix}{delimiter}{wildcard}")

    def get_tag(self) -> str:
        return self.topic

    async def send_message(self, **kwargs: tp.Any) -> None:
        raise NotImplementedError()

    @property
    def result_topic(self) -> str:
        return ""

    async def publish(self, message: "ProtoBunnyMessage") -> None:
        raise NotImplementedError

    async def publish_result(
        self,
        result: "Result",
        topic: str | None = None,
        correlation_id: str | None = None,
    ) -> None:
        raise NotImplementedError

    async def _receive(
        self,
        log_callback: "LoggerCallback",  # the callback function for logging is always a sync function
        message: "IncomingMessageProtocol",
    ) -> None:
        """Call the logging callback.

        Args:
            log_callback: The callback function passed to pb.subscribe_logger().
              It receives the aio_pika IncomingMessage as first argument and the string to log as second.

            message: the aio_pika IncomingMessage
        """
        if message.routing_key is None:
            raise ValueError("Routing key was not set. Invalid topic")
        try:
            body = get_body(message)
            log_callback(message, body)
        except RequeueMessage:
            raise
        except Exception as exc:  # pylint: disable=W0703
            log.exception(
                "Could not process message on Logging queue: %s - %s", str(message.body), str(exc)
            )


def is_task(topic: str) -> bool:
    delimiter = default_configuration.backend_config.topic_delimiter
    return "tasks" in topic.split(delimiter)


# keep always the imports of generated code at the end of the file
from protobunny.core.results import Result, ReturnCode
